# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

__docformat__ = 'reStructuredText'

from copy import copy
import re
import sys
from xml.sax.saxutils import unescape

from reportlab.lib.enums import TA_CENTER, TA_RIGHT
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import cm
from reportlab.platypus.doctemplate import FrameActionFlowable, FrameBreak, Indenter
from reportlab.platypus.flowables import (
    _listWrapOn,
    _FUZZ,
    Flowable,
    NullDraw,
    PageBreak,
    Spacer,
)
from reportlab.platypus.frames import Frame
from reportlab.platypus.paragraph import Paragraph
from reportlab.platypus.tables import Table, TableStyle
from reportlab.platypus.tableofcontents import TableOfContents
from reportlab.platypus.xpreformatted import XPreformatted

from . import styles
from .log import log


class XXPreformatted(XPreformatted):
    """An extended XPreformattedFit"""

    def __init__(self, *args, **kwargs):
        XPreformatted.__init__(self, *args, **kwargs)

    def split(self, aW, aH):

        # Figure out a nice range of splits
        #
        # Assume we would prefer 5 lines (at least) on
        # a splitted flowable before a break, and 4 on
        # the last flowable after a break.
        # So, the minimum wrap height for a fragment
        # will be 5*leading

        rW, rH = self.wrap(aW, aH)
        if rH > aH:

            minH1 = getattr(self.style, 'allowOrphans', 5) * self.style.leading
            minH2 = getattr(self.style, 'allowWidows', 4) * self.style.leading

            # If there's no way to fid a decent fragment,
            # refuse to split
            if aH < minH1:
                return []

            # Now, don't split too close to the end either
            pw, ph = self.wrap(aW, aH)
            if ph - aH < minH2:
                aH = ph - minH2

        return XPreformatted.split(self, aW, aH)


class MyIndenter(Indenter):
    """An indenter that has a width, because otherwise you get crashes
    if added inside tables"""

    width = 0
    height = 0

    def draw(self):
        pass


class TocEntry(NullDraw):
    """A flowable that adds a TOC entry but draws nothing"""

    def __init__(self, level, label):
        self.level = level
        self.label = label
        self.width = 0
        self.height = 0
        self.keepWithNext = True

    def draw(self):
        # Add outline entry
        self.canv.bookmarkHorizontal(self.label, 0, 0 + self.height)
        self.canv.addOutlineEntry(
            self.label, self.label, max(0, int(self.level)), False
        )


class Heading(Paragraph):
    """A paragraph that also adds an outline entry in
    the PDF TOC."""

    def __init__(
        self,
        text,
        style,
        bulletText=None,
        caseSensitive=1,
        level=0,
        snum=None,
        parent_id=None,
        node=None,
        section_header_depth=2,
    ):
        # Issue 114: need to convert "&amp;" to "&" and such.
        # Issue 140: need to make it plain text
        self.stext = re.sub(r'<[^>]*?>', '', unescape(text))
        self.stext = self.stext.strip()
        self.level = int(level)
        self.snum = snum
        self.parent_id = parent_id
        self.node = node
        self.section_header_depth = section_header_depth
        Paragraph.__init__(self, text, style, bulletText)

    def draw(self):

        # Add outline entry
        self.canv.bookmarkHorizontal(self.parent_id, 0, 0 + self.height)
        # self.section_header_depth is for Issue 391
        if self.canv.firstSect and self.level < self.section_header_depth:
            self.canv.sectName = self.stext
            self.canv.firstSect = False
            if self.snum is not None:
                self.canv.sectNum = self.snum
            else:
                self.canv.sectNum = ""
        self.canv.addOutlineEntry(self.stext, self.parent_id, int(self.level), False)
        Paragraph.draw(self)


class Separation(Flowable):
    """A simple <hr>-like flowable"""

    def wrap(self, w, h):
        self.w = w
        return w, 1 * cm

    def draw(self):
        self.canv.line(0, 0.5 * cm, self.w, 0.5 * cm)


class Reference(Flowable):
    """A flowable to insert an anchor without taking space"""

    def __init__(self, refid):
        self.refid = refid
        self.keepWithNext = True
        Flowable.__init__(self)

    def wrap(self, w, h):
        """This takes no space"""
        return 0, 0

    def draw(self):
        self.canv.bookmarkPage(self.refid)

    def repr(self):
        return "Reference: %s" % self.refid

    def __str__(self):
        return "Reference: %s" % self.refid


class OddEven(Flowable):
    """This flowable takes two lists of flowables as arguments, odd and even.
    If will draw the "odd" list when drawn in odd pages and the "even" list on
    even pages.


    wrap() will always return a size large enough for both lists, and this flowable
    **cannot** be split, so use with care.
    """

    def __init__(self, odd, even, style=None):
        self.odd = DelayedTable([[odd]], ['100%'], style)
        self.even = DelayedTable([[even]], ['100%'], style)

    def wrap(self, w, h):
        """Return a box large enough for both odd and even"""
        w1, h1 = self.odd.wrap(w, h)
        w2, h2 = self.even.wrap(w, h)
        return max(w1, w2), max(h1, h2)

    def drawOn(self, canvas, x, y, _sW=0):
        if canvas._pagenum % 2 == 0:
            self.even.drawOn(canvas, x, y, _sW)
        else:
            self.odd.drawOn(canvas, x, y, _sW)

    def split(self):
        """Makes no sense to split this..."""
        return []


class DelayedTable(Table):
    """A flowable that inserts a table for which it has the data.

    Needed so column widths can be determined after we know on what frame
    the table will be inserted, thus making the overal table width correct.

    """

    def __init__(self, data, colWidths, style=None, repeatrows=False, splitByRow=True):
        self.data = data
        self._colWidths = colWidths
        if style is None:
            style = TableStyle(
                [
                    ('LEFTPADDING', (0, 0), (-1, -1), 0),
                    ('RIGHTPADDING', (0, 0), (-1, -1), 0),
                    ('TOPPADDING', (0, 0), (-1, -1), 0),
                    ('BOTTOMPADDING', (0, 0), (-1, -1), 0),
                ]
            )
        self.style = style
        self.t = None
        self.repeatrows = repeatrows
        self.hAlign = TA_CENTER
        self.splitByRow = splitByRow

    def wrap(self, w, h):
        # Create the table, with the widths from colWidths reinterpreted
        # if needed as percentages of frame/cell/whatever width w is.

        # _tw = w/sum(self.colWidths)
        def adjust(*args, **kwargs):
            kwargs['total'] = w
            return styles.adjustUnits(*args, **kwargs)

        # adjust=functools.partial(styles.adjustUnits, total=w)
        self.colWidths = [adjust(x) for x in self._colWidths]
        # colWidths = [_w * _tw for _w in self.colWidths]
        self.t = Table(
            self.data,
            colWidths=self.colWidths,
            style=self.style,
            repeatRows=self.repeatrows,
            splitByRow=True,
        )
        # splitByRow=self.splitByRow)
        self.t.hAlign = self.hAlign
        return self.t.wrap(w, h)

    def split(self, w, h):
        if self.splitByRow:
            if not self.t:
                self.wrap(w, h)
            return self.t.split(w, h)
        else:
            return []

    def drawOn(self, canvas, x, y, _sW=0):
        self.t.drawOn(canvas, x, y, _sW)

    def identity(self, maxLen=None):
        return (
            "<%s at %s%s%s> containing: %s"
            % (
                self.__class__.__name__,
                hex(id(self)),
                self._frameName(),
                getattr(self, 'name', '')
                and (' name="%s"' % getattr(self, 'name', ''))
                or '',
                repr(self.data[0]),
            )[:180]
        )


def tablepadding(padding):
    if not isinstance(padding, (list, tuple)):
        padding = [padding] * 4
    return (
        padding,
        ('TOPPADDING', [0, 0], [-1, -1], padding[0]),
        ('RIGHTPADDING', [-1, 0], [-1, -1], padding[1]),
        ('BOTTOMPADDING', [0, 0], [-1, -1], padding[2]),
        ('LEFTPADDING', [1, 0], [1, -1], padding[3]),
    )


class SplitTable(DelayedTable):
    def __init__(self, data, colWidths, style, padding=3):
        if len(data) != 1 or len(data[0]) != 2:
            log.error('SplitTable can only be 1 row and two columns!')
            sys.exit(1)
        DelayedTable.__init__(self, data, colWidths, style)
        self.padding, p1, p2, p3, p4 = tablepadding(padding)
        self.style._cmds.insert(0, p1)
        self.style._cmds.insert(0, p2)
        self.style._cmds.insert(0, p3)
        self.style._cmds.insert(0, p4)

    def identity(self, maxLen=None):
        return "<%s at %s%s%s> containing: %s" % (
            self.__class__.__name__,
            hex(id(self)),
            self._frameName(),
            getattr(self, 'name', '')
            and (' name="%s"' % getattr(self, 'name', ''))
            or '',
            repr(self.data[0][1])[:180],
        )

    def split(self, w, h):
        _w, _h = self.wrap(w, h)

        if _h > h:  # Can't split!
            # The right column data mandates the split
            # Find which flowable exceeds the available height

            dw = self.colWidths[0] + self.padding[1] + self.padding[3]
            dh = self.padding[0] + self.padding[2]

            bullet = self.data[0][0]
            text = self.data[0][1]
            for l in range(0, len(text)):
                _, fh = _listWrapOn(text[: l + 1], w - dw, None)
                if fh + dh > h:
                    # The lth flowable is the guilty one
                    # split it

                    _, lh = _listWrapOn(text[:l], w - dw, None)
                    # Workaround for Issue 180
                    text[l].wrap(w - dw, h - lh - dh)
                    l2 = text[l].split(w - dw, h - lh - dh)
                    if l2 == []:  # Not splittable, push some to next page
                        if l == 0:  # Can't fit anything, push all to next page
                            return l2

                        # We reduce the number of items we keep on the
                        # page for two reasons:
                        #    1) If an item is associated with the following
                        #       item (getKeepWithNext() == True), we have
                        #       to back up to a previous one.
                        #    2) If we miscalculated the size required on
                        #       the first page (I dunno why, probably not
                        #       counting borders properly, but we do
                        #       miscalculate occasionally).  Seems to
                        #       have to do with nested tables, so it might
                        #       be the extra space on the border on the
                        #       inner table.

                        while l > 0:
                            if not text[l - 1].getKeepWithNext():
                                first_t = Table(
                                    [[bullet, text[:l]]],
                                    colWidths=self.colWidths,
                                    style=self.style,
                                )
                                _w, _h = first_t.wrap(w, h)
                                if _h <= h:
                                    break
                            l -= 1

                        if l > 0:
                            # Workaround for Issue 180 with wordaxe:
                            # if wordaxe is not None:
                            # l3=[Table([
                            # [bullet,
                            # text[:l]]
                            # ],
                            # colWidths=self.colWidths,
                            # style=self.style),
                            # Table([['',text[l:]]],
                            # colWidths=self.colWidths,
                            # style=self.style)]
                            # else:
                            l3 = [
                                first_t,
                                SplitTable(
                                    [['', text[l:]]],
                                    colWidths=self.colWidths,
                                    style=self.style,
                                    padding=self.padding,
                                ),
                            ]
                        else:  # Everything flows
                            l3 = []
                    else:
                        l3 = [
                            Table(
                                [[bullet, text[:l] + [l2[0]]]],
                                colWidths=self.colWidths,
                                rowHeights=[h],
                                style=self.style,
                            )
                        ]
                        if l2[1:] + text[l + 1 :]:
                            l3.append(
                                SplitTable(
                                    [['', l2[1:] + text[l + 1 :]]],
                                    colWidths=self.colWidths,
                                    style=self.style,
                                    padding=self.padding,
                                )
                            )
                    return l3
            log.debug("Can't split splittable")
            return self.t.split(w, h)
        else:
            return DelayedTable.split(self, w, h)


class MySpacer(Spacer):
    def wrap(self, aW, aH):
        w, h = Spacer.wrap(self, aW, aH)
        self.height = min(aH, h)
        return w, self.height


class MyPageBreak(FrameActionFlowable):
    def __init__(self, templateName=None, breakTo='any'):
        """templateName switches the page template starting in the
        next page.

        breakTo can be 'any' 'even' or 'odd'.

        'even' will break one page if the current page is odd
        or two pages if it's even. That way the next flowable
        will be in an even page.

        'odd' is the opposite of 'even'

        'any' is the default, and means it will always break
        only one page.

        """

        self.templateName = templateName
        self.breakTo = breakTo
        self.forced = False
        self.extraContent = []

    def frameAction(self, frame):
        frame._generated_content = []
        if self.breakTo == 'any':  # Break only once. None if at top of page
            if not frame._atTop:
                frame._generated_content.append(SetNextTemplate(self.templateName))
                frame._generated_content.append(PageBreak())
        elif self.breakTo == 'odd':  # Break once if on even page, twice
            # on odd page, none if on top of odd page
            if frame._pagenum % 2:  # odd pageNum
                if not frame._atTop:
                    # Blank pages get no heading or footer
                    frame._generated_content.append(SetNextTemplate(self.templateName))
                    frame._generated_content.append(SetNextTemplate('emptyPage'))
                    frame._generated_content.append(PageBreak())
                    frame._generated_content.append(ResetNextTemplate())
                    frame._generated_content.append(PageBreak())
            else:  # even
                frame._generated_content.append(SetNextTemplate(self.templateName))
                frame._generated_content.append(PageBreak())
        elif self.breakTo == 'even':  # Break once if on odd page, twice
            # on even page, none if on top of even page
            if frame._pagenum % 2:  # odd pageNum
                frame._generated_content.append(SetNextTemplate(self.templateName))
                frame._generated_content.append(PageBreak())
            else:  # even
                if not frame._atTop:
                    # Blank pages get no heading or footer
                    frame._generated_content.append(SetNextTemplate(self.templateName))
                    frame._generated_content.append(SetNextTemplate('emptyPage'))
                    frame._generated_content.append(PageBreak())
                    frame._generated_content.append(ResetNextTemplate())
                    frame._generated_content.append(PageBreak())


class SetNextTemplate(Flowable):
    """Set canv.templateName when drawing.

    rst2pdf uses that to switch page templates.

    """

    def __init__(self, templateName=None):
        self.templateName = templateName
        Flowable.__init__(self)

    def draw(self):
        if self.templateName:
            try:
                self.canv.oldTemplateName = self.canv.templateName
            except Exception:
                self.canv.oldTemplateName = 'oneColumn'
            self.canv.templateName = self.templateName


class ResetNextTemplate(Flowable):
    """Go back to the previous template.

    rst2pdf uses that to switch page templates back when
    temporarily it needed to switch to another template.

    For example, after a OddPageBreak, there can be a totally
    blank page. Those have to use coverPage as a template,
    because they must not have headers or footers.

    And then we need to switch back to whatever was used.

    """

    def __init__(self):
        Flowable.__init__(self)

    def draw(self):
        self.canv.templateName, self.canv.oldTemplateName = (
            self.canv.oldTemplateName,
            self.canv.templateName,
        )

    def wrap(self, aW, aH):
        return 0, 0


class TextAnnotation(Flowable):
    """Add text annotation flowable"""

    def __init__(self, *args):
        self.annotationText = ""
        self.position = [-1, -1, -1, -1]
        if len(args) >= 1:
            self.annotationText = args[0].lstrip('"').rstrip('"')
        if len(args) >= 5:
            self.position = args[1:]

    def wrap(self, w, h):
        return 0, 0

    def draw(self):
        # Format of Reportlab's textAnnotation():
        # textAnnotation("Your content", Rect=[x_begin, y_begin, x_end, y_end], relative=1)
        self.canv.textAnnotation(self.annotationText, self.position, 1)


class Transition(Flowable):
    """Wrap canvas.setPageTransition.

    Sets the transition effect from the current page to the next.

    """

    PageTransitionEffects = dict(
        Split=['direction', 'motion'],
        Blinds=['dimension'],
        Box=['motion'],
        Wipe=['direction'],
        Dissolve=[],
        Glitter=['direction'],
    )

    def __init__(self, *args):
        if len(args) < 1:
            args = [None, 1]  # No transition
        # See if we got a valid transition effect name
        if args[0] not in self.PageTransitionEffects:
            log.error('Unknown transition effect name: %s' % args[0])
            args[0] = None
        elif len(args) == 1:
            args.append(1)

        # FIXME: validate more
        self.args = args

    def wrap(self, aw, ah):
        return 0, 0

    def draw(self):
        kwargs = dict(
            effectname=None, duration=1, direction=0, dimension='H', motion='I'
        )
        ceff = ['effectname', 'duration'] + self.PageTransitionEffects[self.args[0]]
        for argname, argvalue in zip(ceff, self.args):
            kwargs[argname] = argvalue
        kwargs['duration'] = int(kwargs['duration'])
        kwargs['direction'] = int(kwargs['direction'])
        self.canv.setPageTransition(**kwargs)


class SmartFrame(Frame):
    """A (Hopefully) smarter frame object.

    This frame object knows how to handle a two-pass
    layout procedure (someday).

    """

    def __init__(
        self,
        container,
        x1,
        y1,
        width,
        height,
        leftPadding=6,
        bottomPadding=6,
        rightPadding=6,
        topPadding=6,
        id=None,
        showBoundary=0,
        overlapAttachedSpace=None,
        _debug=None,
    ):
        self.container = container
        self.onSidebar = False
        self.__s = '[%s, %s, %s, %s, %s, %s, %s, %s,]' % (
            x1,
            y1,
            width,
            height,
            leftPadding,
            bottomPadding,
            rightPadding,
            topPadding,
        )
        Frame.__init__(
            self,
            x1,
            y1,
            width,
            height,
            leftPadding,
            bottomPadding,
            rightPadding,
            topPadding,
            id,
            showBoundary,
            overlapAttachedSpace,
            _debug,
        )

    def add(self, flowable, canv, trySplit=0):
        flowable._atTop = self._atTop
        return Frame.add(self, flowable, canv, trySplit)

    def __repr__(self):
        return self.__s

    def __deepcopy__(self, *whatever):
        return copy(self)


class FrameCutter(FrameActionFlowable):
    def __init__(self, dx, width, flowable, padding, lpad, floatLeft=True):
        self.width = width
        self.dx = dx
        self.f = flowable
        self.padding = padding
        self.lpad = lpad
        self.floatLeft = floatLeft

    def frameAction(self, frame):
        idx = frame.container.frames.index(frame)
        if self.floatLeft:
            # Don't bother inserting a silly thin frame
            if self.width - self.padding > 30:
                f1 = SmartFrame(
                    frame.container,
                    frame._x1 + self.dx - 2 * self.padding,
                    frame._y2 - self.f.height - 3 * self.padding,
                    self.width + 2 * self.padding,
                    self.f.height + 3 * self.padding,
                    bottomPadding=0,
                    topPadding=0,
                    leftPadding=self.lpad,
                )
                f1._atTop = frame._atTop
                # This is a frame next to a sidebar.
                f1.onSidebar = True
                frame.container.frames.insert(idx + 1, f1)
            # Don't add silly thin frame
            if frame._height - self.f.height - 2 * self.padding > 30:
                frame.container.frames.insert(
                    idx + 2,
                    SmartFrame(
                        frame.container,
                        frame._x1,
                        frame._y1p,
                        self.width + self.dx,
                        frame._height - self.f.height - 3 * self.padding,
                        topPadding=0,
                    ),
                )
        else:
            # Don't bother inserting a silly thin frame
            if self.width - self.padding > 30:
                f1 = SmartFrame(
                    frame.container,
                    frame._x1 - self.width,
                    frame._y2 - self.f.height - 2 * self.padding,
                    self.width,
                    self.f.height + 2 * self.padding,
                    bottomPadding=0,
                    topPadding=0,
                    rightPadding=self.lpad,
                )
                f1._atTop = frame._atTop
                # This is a frame next to a sidebar.
                f1.onSidebar = True
                frame.container.frames.insert(idx + 1, f1)
            if frame._height - self.f.height - 2 * self.padding > 30:
                frame.container.frames.insert(
                    idx + 2,
                    SmartFrame(
                        frame.container,
                        frame._x1 - self.width,
                        frame._y1p,
                        self.width + self.dx,
                        frame._height - self.f.height - 2 * self.padding,
                        topPadding=0,
                    ),
                )


class Sidebar(FrameActionFlowable):
    def __init__(self, flowables, style):
        self.style = style
        self.width = self.style.width
        self.flowables = flowables

    def frameAction(self, frame):
        if self.style.float not in ('left', 'right'):
            return
        if frame.onSidebar:  # We are still on the frame next to a sidebar!
            frame._generated_content = [FrameBreak(), self]
        else:
            w = frame.container.styles.adjustUnits(self.width, frame.width)
            idx = frame.container.frames.index(frame)
            padding = self.style.borderPadding
            width = self.style.width
            self.style.padding = frame.container.styles.adjustUnits(
                str(padding), frame.width
            )
            self.style.width = frame.container.styles.adjustUnits(
                str(width), frame.width
            )
            self.kif = BoxedContainer(self.flowables, self.style)
            if self.style.float == 'left':
                self.style.lpad = frame.leftPadding
                f1 = SmartFrame(
                    frame.container,
                    frame._x1,
                    frame._y1p,
                    w - 2 * self.style.padding,
                    frame._y - frame._y1p,
                    leftPadding=self.style.lpad,
                    rightPadding=0,
                    bottomPadding=0,
                    topPadding=0,
                )
                f1._atTop = frame._atTop
                frame.container.frames.insert(idx + 1, f1)
                frame._generated_content = [
                    FrameBreak(),
                    self.kif,
                    FrameCutter(
                        w,
                        frame.width - w,
                        self.kif,
                        padding,
                        self.style.lpad,
                        True,
                    ),
                    FrameBreak(),
                ]
            elif self.style.float == 'right':
                self.style.lpad = frame.rightPadding
                frame.container.frames.insert(
                    idx + 1,
                    SmartFrame(
                        frame.container,
                        frame._x1 + frame.width - self.style.width,
                        frame._y1p,
                        w,
                        frame._y - frame._y1p,
                        rightPadding=self.style.lpad,
                        leftPadding=0,
                        bottomPadding=0,
                        topPadding=0,
                    ),
                )
                frame._generated_content = [
                    FrameBreak(),
                    self.kif,
                    FrameCutter(
                        w,
                        frame.width - w,
                        self.kif,
                        padding,
                        self.style.lpad,
                        False,
                    ),
                    FrameBreak(),
                ]


class BoundByWidth(Flowable):
    """Limit a list of flowables by width.

    This still lets the flowables break over pages and frames.

    """

    def __init__(self, maxWidth, content=[], style=None, mode=None, scale=None):
        self.maxWidth = maxWidth
        self.content = content
        self.style = style
        self.mode = mode
        self.pad = None
        self.scale = scale
        Flowable.__init__(self)

    def border_padding(self, useWidth, additional):
        sdict = self.style
        sdict = sdict.__dict__ or {}
        bp = sdict.get("borderPadding", 0)
        if useWidth:
            additional += sdict.get("borderWidth", 0)
        if not isinstance(bp, list):
            bp = [bp] * 4
        return [x + additional for x in bp]

    def identity(self, maxLen=None):
        return "<%s at %s%s%s> containing: %s" % (
            self.__class__.__name__,
            hex(id(self)),
            self._frameName(),
            getattr(self, 'name', '')
            and (' name="%s"' % getattr(self, 'name', ''))
            or '',
            repr([c.identity() for c in self.content])[:80],
        )

    def wrap(self, availWidth, availHeight):
        """If we need more width than we have, complain, keep a scale"""
        self.pad = self.border_padding(True, 0.1)
        maxWidth = float(
            min(
                styles.adjustUnits(self.maxWidth, availWidth) or availWidth,
                availWidth,
            )
        )
        self.maxWidth = maxWidth
        maxWidth -= self.pad[1] + self.pad[3]
        self.width, self.height = _listWrapOn(
            self.content, maxWidth, None, fakeWidth=False
        )
        if self.width > maxWidth:
            if self.mode != 'shrink':
                self.scale = 1.0
                log.warning(
                    "BoundByWidth too wide to fit in frame (%s > %s): %s",
                    self.width,
                    maxWidth,
                    self.identity(),
                )
            if self.mode == 'shrink' and not self.scale:
                self.scale = (maxWidth + self.pad[1] + self.pad[3]) / (
                    self.width + self.pad[1] + self.pad[3]
                )
        else:
            self.scale = 1.0
        self.height *= self.scale
        self.width *= self.scale
        return (
            self.width,
            self.height + (self.pad[0] + self.pad[2]) * self.scale,
        )

    def split(self, availWidth, availHeight):
        if not self.pad:
            self.wrap(availWidth, availHeight)
        content = self.content
        if len(self.content) == 1:
            # We need to split the only element we have
            content = content[0].split(
                availWidth - (self.pad[1] + self.pad[3]),
                availHeight - (self.pad[0] + self.pad[2]),
            )
        result = [
            BoundByWidth(self.maxWidth, [f], self.style, self.mode, self.scale)
            for f in content
        ]
        return result

    def draw(self):
        """we simulate being added to a frame"""
        canv = self.canv
        canv.saveState()
        x = canv._x
        y = canv._y
        _sW = 0
        scale = self.scale
        content = None
        # , canv, x, y, _sW=0, scale=1.0, content=None, aW=None):
        pS = 0
        aW = self.width
        aW = scale * (aW + _sW)
        if content is None:
            content = self.content
        y += (self.height + self.pad[2]) / scale
        x += self.pad[3]
        for c in content:
            w, h = c.wrapOn(canv, aW, 0xFFFFFFF)
            if (w < _FUZZ or h < _FUZZ) and not getattr(c, '_ZEROSIZE', None):
                continue
            if c is not content[0]:
                h += max(c.getSpaceBefore() - pS, 0)
            y -= h
            canv.saveState()
            if self.mode == 'shrink':
                canv.scale(scale, scale)
            elif self.mode == 'truncate':
                p = canv.beginPath()
                p.rect(
                    x - self.pad[3],
                    y - self.pad[2],
                    self.maxWidth,
                    self.height + self.pad[0] + self.pad[2],
                )
                canv.clipPath(p, stroke=0)
            c.drawOn(canv, x, y, _sW=aW - w)
            canv.restoreState()
            if c is not content[-1]:
                pS = c.getSpaceAfter()
                y -= pS
        canv.restoreState()


class BoxedContainer(BoundByWidth):
    def __init__(self, content, style, mode='shrink'):
        try:
            w = style.width
        except AttributeError:
            w = '100%'
        BoundByWidth.__init__(self, w, content, mode=mode, style=None)
        self.style = style
        self.mode = mode

    def identity(self, maxLen=None):
        return repr(
            ['BoxedContainer containing: ', [c.identity() for c in self.content]]
        )[:80]

    def draw(self):
        canv = self.canv
        canv.saveState()
        x = canv._x
        y = canv._y
        lw = 0
        if self.style and self.style.borderWidth > 0:
            lw = self.style.borderWidth
            canv.setLineWidth(self.style.borderWidth)
            if self.style.borderColor:  # This could be None :-(
                canv.setStrokeColor(self.style.borderColor)
                stroke = 1
            else:
                stroke = 0
        else:
            stroke = 0
        if self.style and self.style.backColor:
            canv.setFillColor(self.style.backColor)
            fill = 1
        else:
            fill = 0

        padding = self.border_padding(False, lw)
        xpadding = padding[1] + padding[3]
        ypadding = padding[0] + padding[2]
        p = canv.beginPath()
        p.rect(x, y, self.width + xpadding, self.height + ypadding)
        canv.drawPath(p, stroke=stroke, fill=fill)
        canv.restoreState()
        BoundByWidth.draw(self)

    def split(self, availWidth, availHeight):
        self.wrap(availWidth, availHeight)
        padding = (self.pad[1] + self.pad[3]) * self.scale
        if self.height + padding <= availHeight:
            return [self]
        else:
            # Try to figure out how many elements
            # we can put in the available space
            candidate = None
            remainder = None
            for p in range(1, len(self.content)):
                b = BoxedContainer(self.content[:p], self.style, self.mode)
                w, h = b.wrap(availWidth, availHeight)
                if h < availHeight:
                    candidate = b
                    if self.content[p:]:
                        remainder = BoxedContainer(
                            self.content[p:], self.style, self.mode
                        )
                else:
                    break
            if not candidate or not remainder:  # Nothing fits, break page
                return []
            if not remainder:  # Everything fits?
                return [self]
            return [candidate, remainder]


class MyTableOfContents(TableOfContents):
    """
    Subclass of reportlab.platypus.tableofcontents.TableOfContents
    which supports hyperlinks to corresponding sections.
    """

    def __init__(self, *args, **kwargs):

        # The parent argument is to define the locality of
        # the TOC. If it's none, it's a global TOC and
        # any heading it's notified about is accepted.

        # If it's a node, then the heading needs to be "inside"
        # that node. This can be figured out because
        # the heading flowable keeps a reference to the title
        # node it was creatd from.
        #
        # Yes, this is gross.

        self.parent = kwargs.pop('parent')
        TableOfContents.__init__(self, *args, **kwargs)
        # reference ids for which this TOC should be notified
        self.refids = []
        # revese lookup table from (level, text) to refid
        self.refid_lut = {}
        self.linkColor = "#0000ff"

    def notify(self, kind, stuff):
        # stuff includes (level, text, pagenum, label)
        level, text, pageNum, label, node = stuff
        rlabel = '-'.join(label.split('-')[:-1])

        def islocal(_node):
            """See if this node is "local enough" for this TOC.
            This is for Issue 196"""
            if self.parent is None:
                return True
            while _node.parent:
                if _node.parent == self.parent:
                    return True
                _node = _node.parent
            return False

        if rlabel in self.refids and islocal(node):
            self.addEntry(level, text, pageNum)
            self.refid_lut[(level, text, pageNum)] = label

    def wrap(self, availWidth, availHeight):
        """Adds hyperlink to toc entry."""

        widths = (availWidth - self.rightColumnWidth, self.rightColumnWidth)

        # makes an internal table which does all the work.
        # we draw the LAST RUN's entries!  If there are
        # none, we make some dummy data to keep the table
        # from complaining
        if len(self._lastEntries) == 0:
            _tempEntries = [(0, 'Placeholder for table of contents', 0, None)]
        else:
            _tempEntries = self._lastEntries

        if _tempEntries:
            base_level = _tempEntries[0][0]
        else:
            base_level = 0
        tableData = []
        for entry in _tempEntries:
            level, text, pageNum = entry[:3]
            left_col_level = level - base_level
            leftColStyle = self.getLevelStyle(left_col_level)
            label = self.refid_lut.get((level, text, pageNum), None)
            if label:
                pre = u'<a href="#%s" color="%s">' % (label, self.linkColor)
                post = u'</a>'
                if isinstance(text, bytes):
                    text = text.decode('utf-8')
                text = pre + text + post
            else:
                pre = ''
                post = ''
            # right col style is right aligned
            rightColStyle = ParagraphStyle(
                name='leftColLevel%d' % left_col_level,
                parent=leftColStyle,
                leftIndent=0,
                alignment=TA_RIGHT,
            )
            leftPara = Paragraph(text, leftColStyle)
            rightPara = Paragraph(pre + str(pageNum) + post, rightColStyle)
            tableData.append([leftPara, rightPara])

        self._table = Table(tableData, colWidths=widths, style=self.tableStyle)

        self.width, self.height = self._table.wrapOn(self.canv, availWidth, availHeight)
        return self.width, self.height

    def split(self, aW, aH):
        # Make sure _table exists before splitting.
        # This was only triggered in rare cases using sphinx.
        if not self._table:
            self.wrap(aW, aH)
        return TableOfContents.split(self, aW, aH)

    def isSatisfied(self):
        if self._entries == self._lastEntries:
            log.debug('Table Of Contents is stable')
            return True
        else:
            if len(self._entries) != len(self._lastEntries):
                log.info(
                    'Number of items in TOC changed '
                    'from %d to %d, not satisfied'
                    % (len(self._lastEntries), len(self._entries))
                )
                return False

            log.info('TOC entries that moved in this pass:')
            for i in range(len(self._entries)):
                if self._entries[i] != self._lastEntries[i]:
                    log.info(str(self._entries[i]))
                    log.info(str(self._lastEntries[i]))

        return False



'''
A rst2pdf extension to implement something similar to sphinx's plantuml extension
(see http://pypi.python.org/pypi/sphinxcontrib-plantuml)

Therefore, stuff may be copied from that code.
Ergo:

    :copyright: Copyright 2010 by Yuya Nishihara <yuya@tcha.org>.
    :license: BSD, (he says see LICENSE but the file is not there ;-)

'''

import errno
import subprocess
import tempfile

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives

import rst2pdf.genelements as genelements
from rst2pdf.image import MyImage
from rst2pdf.styles import adjustUnits


class plantuml(nodes.General, nodes.Element):
    pass


class UmlDirective(rst.Directive):
    """Directive to insert PlantUML markup

    Example::

        .. uml::
           :alt: Alice and Bob

           Alice -> Bob: Hello
           Alice <- Bob: Hi


    You can use a :format: option to change between SVG and PNG diagrams, however,
    the SVG plantuml generates doesn't look very good to me.

    Also, :width: and :height: are supported as per the image directive.
    """

    has_content = True
    option_spec = {
        'alt': directives.unchanged,
        'format': directives.unchanged,
        'width': directives.length_or_unitless,
        'height': directives.length_or_unitless,
    }

    def run(self):
        node = plantuml()
        node['uml'] = '\n'.join(self.content)
        node['alt'] = self.options.get('alt', None)
        node['format'] = self.options.get('format', 'png')
        node['width'] = self.options.get('width', None)
        node['height'] = self.options.get('height', None)
        return [node]


class PlantUmlError(Exception):
    pass


class UMLHandler(genelements.NodeHandler, plantuml):
    """Class to handle UML nodes"""

    def gather_elements(self, client, node, style):
        # Create image calling plantuml
        tfile = tempfile.NamedTemporaryFile(
            dir='.', delete=False, suffix='.' + node['format']
        )
        args = 'plantuml -pipe -charset utf-8'
        if node['format'].lower() == 'svg':
            args += ' -tsvg'
        client.to_unlink.append(tfile.name)
        try:
            p = subprocess.Popen(
                args.split(),
                stdout=tfile,
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except OSError as err:
            if err.errno != errno.ENOENT:
                raise
            raise PlantUmlError(
                'plantuml command %r cannot be run' % self.builder.config.plantuml
            )
        serr = p.communicate(node['uml'].encode('utf-8'))[1]
        if p.returncode != 0:
            raise PlantUmlError('error while running plantuml\n\n' + serr)

        # Convert width and height if necessary
        w = node['width']
        if w is not None:
            w = adjustUnits(w)

        h = node['height']
        if h is not None:
            h = adjustUnits(h)

        # Add Image node with the right image
        return [MyImage(tfile.name, client=client, width=w, height=h)]


directives.register_directive("uml", UmlDirective)



# -*- coding: utf-8 -*-

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'rst2pdf.pdfbuilder']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Test Case'
copyright = u'2009, vorsorge'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '1'
# The full version, including alpha/beta/rc tags.
release = '1'

# -- Options for PDF output ----------------------------------------------------

# Grouping the document tree into PDF files. List of tuples
# (source start file, target name, title, author, options).
pdf_documents = [('index', 'TestCase', u'A Sphinx Test Case', u'vorsorge')]

# A comma-separated list of custom stylesheets. Example:
pdf_stylesheets = ['sphinx', 'letter']

# Section level that forces a break page.
# For example: 1 means top-level sections start in a new page
# 0 means disabled
pdf_break_level = 1

# verbosity level. 0 1 or 2
pdf_verbosity = 2

# If false, no coverpage is generated.
pdf_use_coverpage = True

pdf_invariant = True
pdf_real_footnotes = True

# Set a consistent date for the cover page
today = 'April 29, 2018'


#!/usr/bin/env python
'''
    Call dumps() to dump a stylesheet to a string.

    Or run the script to dump all .json in the styles directory
    to .style in the styles directory.
'''

import sys
import os

from rst2pdf.rson import loads as rloads
from json import loads as jloads


def dumps(obj, forcestyledict=True):
    """If forcestyledict is True, will attempt to
    turn styles into a dictionary.
    """

    def dofloat(result, obj, indent):
        s = '%.3f' % obj
        while '.' in s and s.endswith('0'):
            s = s[:-1]
        result.append(s)

    def doint(result, obj, indent):
        if isinstance(obj, bool):
            obj = repr(obj).lower()
        result.append(str(obj))

    badch = set('[]{}:=,"\n')

    def dostr(result, obj, indent):
        try:
            float(obj)
        except ValueError:
            ok = True
        else:
            ok = obj == obj.strip()
        ok = ok and not set(obj) & badch
        if ok:
            result.append(obj)
            return

        obj = obj.replace('\\', '\\\\').replace('\n', '\\n')
        obj = obj.replace('"', '\\"')
        result.append('"%s"' % obj)

    def dolist(result, obj, indent):
        if indent:
            if not obj:
                result.append('[]')
                return
            elif isinstance(obj[0], list):
                result.append('[]')
                indent += '    '
                for item in obj:
                    dumprecurse(result, item, indent)
                return
        result.append('[')
        obj = [[x, ', '] for x in obj]
        obj[-1][-1] = ']'
        for item, separator in obj:
            dumprecurse(result, item, '')
            result.append(separator)

    def dodict(result, obj, indent):
        if not obj:
            result.append('{}')
            return
        obj = sorted(obj.items())
        multiline = indent and (
            len(obj) > 2
            or len(obj) == 2
            and (
                isinstance(obj[0][-1], (list, dict))
                or isinstance(obj[-1][-1], (list, dict))
            )
        )
        if not multiline and (not indent or len(obj) != 1):
            result.append('{')
            obj = [[x, ', '] for x in obj]
            obj[-1][-1] = '}'
            for (key, value), separator in obj:
                dumprecurse(result, key, '')
                result.append(': ')
                dumprecurse(result, value, '')
                result.append(separator)
            return
        doindent = len(obj) > 1
        for key, value in obj:
            dumprecurse(result, key, indent, doindent)
            result.append(': ')
            dumprecurse(result, value, indent + '    ', False)

    def donone(result, obj, indent):
        result.append('null')

    dumpfuncs = {
        float: dofloat,
        int: doint,
        str: dostr,
        list: dolist,
        dict: dodict,
        type(None): donone,
    }

    dumpfuncs = dumpfuncs.items()

    def dumprecurse(result, obj, indent='\n', indentnow=True):
        if indentnow:
            result.append(indent)
        for otype, ofunc in dumpfuncs:
            if isinstance(obj, otype):
                return ofunc(result, obj, indent)
        raise ValueError(repr(obj))

    result = []
    if forcestyledict:
        obj = fixstyle(obj)
    dumprecurse(result, obj, indentnow=False)
    return fixspacing(''.join(result))


def fixspacing(s):
    """Try to make the output prettier by inserting blank lines
    in random places.
    """
    result = []
    indent = -1
    for line in s.splitlines():
        line = line.rstrip()  # Some lines had ' '
        if not line:
            continue
        indent, previndent = len(line) - len(line.lstrip()), indent
        if indent <= previndent and indent < 8:
            if indent < previndent or not indent:
                result.append('')
        result.append(line)
    result.append('')
    return '\n'.join(result)


def fixstyle(obj):
    """Try to convert styles into a dictionary"""
    if obj:
        if isinstance(obj, list):
            lengths = [len(x) for x in obj]
            if min(lengths) == max(lengths) == 2:
                obj = dict(obj)
        elif isinstance(obj, dict) and 'styles' in obj:
            obj['styles'] = dict(obj['styles'])
    return obj


def convert(srcname):
    """Convert a single file from .json to .style"""
    print(srcname)
    sstr = open(srcname, 'rb').read()
    sdata = fixstyle(jloads(sstr))
    dstr = dumps(sdata)
    assert sdata == rloads(dstr), "Bad round-trip"

    dstname = srcname.replace('.json', '.style')
    dstf = open(dstname, 'wb')
    dstf.write(dstr)
    dstf.close()


if __name__ == '__main__':
    _dir = os.path.dirname(sys.argv[0])
    _stylesdir = os.path.join(_dir, 'styles')
    for fname in [
        os.path.join('styles', x) for x in os.listdir(_stylesdir) if x.endswith('.json')
    ]:
        convert(fname)



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

'''
This module contains sphinx-specific node handlers.  An import
of this module will apparently fail if sphinx.roles hasn't been
imported.

This module creates a sphinx-specific dispatch dictionary,
which is kept separate from the regular one.

When the SphinxHandler class is instantiated, the two dictionaries
are combined into the instantiated object.
'''

from copy import copy

import docutils
from reportlab.platypus import Paragraph, TableStyle
import sphinx

from .flowables import (
    MySpacer,
    MyIndenter,
    DelayedTable,
)
from .image import MyImage
from .log import nodeid, log
from .nodehandlers import NodeHandler, FontHandler, HandleEmphasis

# ################# NodeHandler subclasses ###################


class SphinxHandler(NodeHandler):
    sphinxmode = True
    dispatchdict = {}

    def __init__(self):
        """This is where the magic happens.  Make a copy of the elements
        in the non-sphinx dispatch dictionary, setting sphinxmode on
        every element, and then overwrite that dictionary with any
        sphinx-specific handlers.
        """
        mydict = {}
        for key, value in self._baseclass.dispatchdict.items():
            value = copy(value)
            value.sphinxmode = True
            mydict[key] = value
        mydict.update(self.dispatchdict)
        self.dispatchdict = mydict


class SphinxFont(SphinxHandler, FontHandler):
    pass


class HandleSphinxDefaults(
    SphinxHandler,
    sphinx.addnodes.glossary,
    sphinx.addnodes.start_of_file,
    sphinx.addnodes.compact_paragraph,
    sphinx.addnodes.pending_xref,
):
    pass


class SphinxListHandler(SphinxHandler):
    def get_text(self, client, node, replaceEnt):
        t = client.gather_pdftext(node)
        while t and t[0] in ', ':
            t = t[1:]
        return t


class HandleSphinxDescAddname(SphinxFont, sphinx.addnodes.desc_addname):
    fontstyle = "descclassname"


class HandleSphinxDescName(SphinxFont, sphinx.addnodes.desc_name):
    fontstyle = "descname"


class HandleSphinxDescReturn(SphinxFont, sphinx.addnodes.desc_returns):
    def get_font_prefix(self, client, node, replaceEnt):
        return ' &rarr; ' + client.styleToFont("returns")


class HandleSphinxDescType(SphinxFont, sphinx.addnodes.desc_type):
    fontstyle = "desctype"


class HandleSphinxDescParamList(SphinxListHandler, sphinx.addnodes.desc_parameterlist):
    pre = ' ('
    post = ')'


class HandleSphinxDescParam(SphinxFont, sphinx.addnodes.desc_parameter):
    fontstyle = "descparameter"

    def get_pre_post(self, client, node, replaceEnt):
        pre, post = FontHandler.get_pre_post(self, client, node, replaceEnt)
        if node.hasattr('noemph'):
            pre = ', ' + pre
        else:
            pre = ', <i>' + pre
            post += '</i>'
        return pre, post


class HandleSphinxDescOpt(SphinxListHandler, SphinxFont, sphinx.addnodes.desc_optional):
    fontstyle = "optional"

    def get_pre_post(self, client, node, replaceEnt):
        prepost = FontHandler.get_pre_post(self, client, node, replaceEnt)
        return '%s[%s, ' % prepost, '%s]%s' % prepost


class HandleDescAnnotation(
    SphinxHandler, HandleEmphasis, sphinx.addnodes.desc_annotation
):
    pass


class HandleSphinxIndex(SphinxHandler, sphinx.addnodes.index):
    def gather_elements(self, client, node, style):
        try:
            for entry in node['entries']:
                client.pending_targets.append(docutils.nodes.make_id(entry[2]))
        except IndexError:
            if node['entries']:
                log.error(
                    "Can't process index entry: %s [%s]",
                    node['entries'],
                    nodeid(node),
                )
        return []


# custom SPHINX nodes.
# FIXME: make sure they are all here, and keep them all together


class HandleSphinxCentered(SphinxHandler, sphinx.addnodes.centered):
    def gather_elements(self, client, node, style):
        return [Paragraph(client.gather_pdftext(node), client.styles['centered'])]


class HandleSphinxDesc(SphinxHandler, sphinx.addnodes.desc):
    def gather_elements(self, client, node, style):
        # Most node['desctype'] do not have a style attached and the default of "normal"
        # is acceptable, so suppress the warning as it's irrelevant in this situation
        client.styles.suppress_undefined_style_warning = True
        st = client.styles[node['desctype']]
        client.styles.suppress_undefined_style_warning = False
        if st == client.styles['normal']:
            st = copy(client.styles['desc'])
            st.spaceBefore = 0
        pre = [MySpacer(0, client.styles['desc'].spaceBefore)]
        return pre + client.gather_elements(node, st)


class HandleSphinxDescSignature(SphinxHandler, sphinx.addnodes.desc_signature):
    def gather_elements(self, client, node, style):
        # Need to add ids as targets, found this when using one of the
        # django docs extensions
        targets = [i.replace(' ', '') for i in node['ids']]
        pre = ''
        for i in targets:
            if i not in client.targets:
                pre += '<a name="%s" />' % i
                client.targets.append(i)
        return [Paragraph(pre + client.gather_pdftext(node), style)]


class HandleSphinxDescSignatureLine(SphinxHandler, sphinx.addnodes.desc_signature_line):
    def gather_elements(self, client, node, style):
        pass


class HandleSphinxDescContent(SphinxHandler, sphinx.addnodes.desc_content):
    def gather_elements(self, client, node, style):
        return (
            [MyIndenter(left=10)]
            + client.gather_elements(node, client.styles["definition"])
            + [MyIndenter(left=-10)]
        )


class HandleHList(SphinxHandler, sphinx.addnodes.hlist):
    def gather_elements(self, client, node, style):
        # Each child is a hlistcol and represents a column.
        # Each grandchild is a bullet list that's the contents
        # of the column

        # Represent it as a N-column, 1-row table, each cell containing
        # a list.

        cells = [[client.gather_elements(child, style) for child in node.children]]
        t_style = TableStyle(client.styles['hlist'].commands)
        cw = 100.0 / len(node.children)
        return [
            DelayedTable(cells, colWidths=['%s%%' % cw] * len(cells), style=t_style)
        ]


class HandleHighlightLang(SphinxHandler, sphinx.addnodes.highlightlang):
    pass


try:
    x = sphinx.ext.graphviz.graphviz

    class HandleSphinxGraphviz(SphinxHandler, sphinx.ext.graphviz.graphviz):
        def gather_elements(self, client, node, style):
            # Based on the graphviz extension
            try:
                # Use bitmap
                fname, outfn = sphinx.ext.graphviz.render_dot(
                    node['builder'], node['code'], node['options'], 'png'
                )
                if outfn:
                    client.to_unlink.append(outfn)
                    client.to_unlink.append(outfn + '.map')
                else:
                    # Something went very wrong with graphviz, and
                    # sphinx should have given an error already
                    return []
            except sphinx.ext.graphviz.GraphvizError as exc:
                log.error('dot code %r: ' % node['code'] + str(exc))
                return [Paragraph(node['code'], client.styles['code'])]
            return [MyImage(filename=outfn, client=client)]

except AttributeError:
    # Probably the graphviz extension is not enabled
    pass


sphinxhandlers = SphinxHandler()



# -*- coding: utf-8 -*-

from docutils.nodes import Text, target
from docutils.parsers.rst import roles

values = {}


class CounterNode(Text):
    children = ()

    def __init__(self, data, rawsource=''):
        if ':' in data:
            self.name, value = [s.lower() for s in data.split(':')][:2]
            self.value = int(value)
        else:
            self.name = data.lower()
            self.value = values.get(self.name, 1)
        values[self.name] = self.value + 1

    def astext(self):
        return str(self.value)


def counter_fn(name, rawtext, text, lineno, inliner, options={}, content=[]):
    n = CounterNode(text)
    s = '%s-%s' % (n.name, n.value)
    return [target(ids=[s]), n], []


counter_fn.content = True

roles.register_canonical_role('counter', counter_fn)



# SPDX-License-Identifier: MIT

from docutils import nodes
from docutils.parsers.rst import roles
import importlib.metadata
import packaging.version


def _get_version(package):
    """Get version from the usual methods."""
    try:
        parsed_version = packaging.version.parse(importlib.metadata.version(package))
    except importlib.metadata.PackageNotFoundError:
        return 'UNKNOWN', 'UNKNOWN'

    version = '.'.join(str(r) for r in parsed_version.release)

    if parsed_version.pre:
        revision = parsed_version.pre
    elif parsed_version.dev:
        revision = f'dev{parsed_version.dev}'
    elif parsed_version.is_postrelease:
        revision = f'post{parsed_version.post}'
    else:
        revision = 'final'

    return version, revision


def version_fn(name, rawtext, text, lineno, inliner, options=None, content=None):
    version, _ = _get_version(text)
    return [nodes.Text(version)], []


def revision_fn(name, rawtext, text, lineno, inliner, options=None, content=None):
    _, revision = _get_version(text)
    return [nodes.Text(revision)], []


roles.register_canonical_role('package-version', version_fn)
roles.register_canonical_role('package-revision', revision_fn)



#!/usr/bin/env python3
# -*- coding: utf8 -*-
# :Copyright: © 2015 Günter Milde.
# :License: Released under the terms of the `2-Clause BSD license`_, in short:
#
#    Copying and distribution of this file, with or without modification,
#    are permitted in any medium without royalty provided the copyright
#    notice and this notice are preserved.
#    This file is offered as-is, without any warranty.
#
# .. _2-Clause BSD license: http://www.spdx.org/licenses/BSD-2-Clause

"""
A version of rst2html5 with support for rst2pdf's custom roles/directives.
"""

import locale

from docutils.core import default_description
from docutils.core import publish_cmdline
from docutils.parsers.rst import directives

from rst2pdf.directives import code_block
from rst2pdf.directives import noop
from rst2pdf.roles import counter_off  # noqa
from rst2pdf.roles import package  # noqa


locale.setlocale(locale.LC_ALL, '')

directives.register_directive('code-block', code_block.code_block_directive)
directives.register_directive('oddeven', noop.noop_directive)

description = (
    'Generates HTML5 documents from standalone reStructuredText '
    'sources.\n' + default_description
)

publish_cmdline(writer_name='html5', description=description)



# -*- coding: utf-8 -*-

from docutils.parsers.rst import roles


def counter_fn(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return [], []


counter_fn.content = False

roles.register_canonical_role('counter', counter_fn)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

"""
Scan a list of folders and find all .afm files,
then create rst2pdf-ready font-aliases.
"""

import os
import subprocess
import sys
from fnmatch import fnmatch

from reportlab.lib.fonts import addMapping
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import (
    FF_FORCEBOLD,
    FF_ITALIC,
    TTFError,
    TTFont,
    TTFontFile,
)

from rst2pdf.log import log

flist = []
afmList = []
pfbList = {}
ttfList = []

# Aliases defined by GhostScript, so if you use Palatino or whatever you
# may get **something**. They are family name aliases.
Alias = {
    "itc bookman": "urw bookman l",
    "itc avant garde gothic": "urw gothic l",
    "palatino": "urw palladio l",
    "new century schoolbook": "century schoolbook l",
    "itc zapf chancery": "urw chancery l",
}

# Standard PDF fonts, so no need to embed them
Ignored = ["times", "itc zapf dingbats", "symbol", "helvetica", "courier"]


fonts = {}
families = {}
fontMappings = {}


def make_string(b):
    if isinstance(b, bytes):
        return b.decode('utf-8')
    return b


def loadFonts():
    """
    Search the system and build lists of available fonts.
    """

    if not any([afmList, pfbList, ttfList]):
        # FIXME: When we drop support por py2, use recursive globs
        for folder in flist:
            for root, _, files in os.walk(folder):
                for f in files:
                    if fnmatch(f, "*.ttf") or fnmatch(f, "*.ttc"):
                        ttfList.append(os.path.join(root, f))
                    elif fnmatch(f, "*.afm"):
                        afmList.append(os.path.join(root, f))
                    elif fnmatch(f, "*.pfb"):
                        pfbList[f[:-4]] = os.path.join(root, f)

        for ttf in ttfList:
            try:
                font = TTFontFile(ttf)
            except TTFError:
                log.warning("Error processing %s", ttf)
                continue

            family = make_string(font.familyName.lower())
            fontName = make_string(font.name).lower()
            baseName = os.path.basename(ttf)[:-4]
            fullName = make_string(font.fullName).lower()

            for k in (fontName, fullName, fullName.replace("italic", "oblique")):
                fonts[k] = (ttf, ttf, family)

            bold = FF_FORCEBOLD == FF_FORCEBOLD & font.flags
            italic = FF_ITALIC == FF_ITALIC & font.flags

            # And we can try to build/fill the family mapping
            if family not in families:
                families[family] = [fontName, fontName, fontName, fontName]
            if bold and italic:
                families[family][3] = fontName
            elif bold:
                families[family][1] = fontName
            elif italic:
                families[family][2] = fontName
            # FIXME: what happens if there are Demi and Medium
            # weights? We get a random one.
            else:
                families[family][0] = fontName

        # Now we have full afm and pbf lists, process the
        # afm list to figure out family name, weight and if
        # it's italic or not, as well as where the
        # matching pfb file is

        for afm in afmList:
            family = None
            fontName = None
            italic = False
            bold = False
            for line in open(afm, "r"):
                line = line.strip()
                if line.startswith("StartCharMetrics"):
                    break
                elif line.startswith("FamilyName"):
                    family = line.split(" ", 1)[1].lower()
                elif line.startswith("FontName"):
                    fontName = line.split(" ")[1]
                elif line.startswith("FullName"):
                    fullName = line.split(" ", 1)[1]
                elif line.startswith("Weight"):
                    bold = line.split(" ")[1] == "Bold"
                elif line.startswith("ItalicAngle"):
                    italic = line.split(" ")[1] != "0.0"

            baseName = os.path.basename(afm)[:-4]
            if family in Ignored or family in Alias:
                continue
            if baseName not in pfbList:
                log.info("afm file without matching pfb file: %s" % baseName)
                continue

            # So now we have a font we know we can embed.
            for n in (
                fontName.lower(),
                fullName.lower(),
                fullName.lower().replace("italic", "oblique"),
            ):
                fonts[n] = (afm, pfbList[baseName], family)

            # And we can try to build/fill the family mapping
            if family not in families:
                families[family] = [fontName, fontName, fontName, fontName]
            if bold and italic:
                families[family][3] = fontName
            elif bold:
                families[family][1] = fontName
            elif italic:
                families[family][2] = fontName
            # FIXME: what happens if there are Demi and Medium
            # weights? We get a random one.
            else:
                families[family][0] = fontName


def findFont(fname):
    loadFonts()
    # So now we are sure we know the families and font
    # names. Well, return some data!
    fname = fname.lower()
    if fname in fonts:
        font = fonts[fname.lower()]
    else:
        if fname in Alias:
            fname = Alias[fname]
        if fname in families:
            font = fonts[families[fname][0].lower()]
        else:
            return None
    return font


def findTTFont(fname):
    def get_family(query):
        data = make_string(subprocess.check_output(["fc-match", query]))
        for line in data.splitlines():
            line = line.strip()
            if not line:
                continue
            family = line.split('"')[1].replace('"', "")
            if family:
                return family
        return None

    def get_fname(query):
        data = make_string(subprocess.check_output(["fc-match", "-v", query]))
        for line in data.splitlines():
            line = line.strip()
            if line.startswith("file: "):
                return line.split('"')[1]
        return None

    def get_variants(family):
        variants = [
            get_fname(family + ":style=Roman"),
            get_fname(family + ":style=Bold"),
            get_fname(family + ":style=Oblique"),
            get_fname(family + ":style=Bold Oblique"),
        ]
        if variants[2] == variants[0]:
            variants[2] = get_fname(family + ":style=Italic")
        if variants[3] == variants[0]:
            variants[3] = get_fname(family + ":style=Bold Italic")
        if variants[0].endswith(".pfb") or variants[0].endswith(".gz"):
            return None
        return variants

    if os.name != "nt":
        family = get_family(fname)
        if not family:
            log.error("Unknown font: %s", fname)
            return None
        return get_variants(family)
    else:
        # lookup required font in registry lookup, alternative approach
        # is to let loadFont() traverse windows font directory or use
        # ctypes with EnumFontFamiliesEx

        def get_nt_fname(ftname):
            import winreg as _w

            fontkey = _w.OpenKey(
                _w.HKEY_LOCAL_MACHINE,
                r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts",
            )
            fontname = ftname + " (TrueType)"
            try:
                fname = _w.QueryValueEx(fontkey, fontname)[0]
                if os.path.isabs(fname):
                    fontkey.close()
                    return fname
                fontdir = os.environ.get("SystemRoot", "C:\\Windows")
                fontdir += "\\Fonts"
                fontkey.Close()
                return fontdir + "\\" + fname
            except WindowsError:
                fontkey.Close()
                return None

        family, pos = guessFont(fname)
        fontfile = get_nt_fname(fname)
        if not fontfile:
            if pos == 0:
                fontfile = get_nt_fname(family)
            elif pos == 1:
                fontfile = get_nt_fname(family + " Bold")
            elif pos == 2:
                fontfile = get_nt_fname(family + " Italic") or get_nt_fname(
                    family + " Oblique"
                )
            else:
                fontfile = get_nt_fname(family + " Bold Italic") or get_nt_fname(
                    family + " Bold Oblique"
                )

            if not fontfile:
                log.error("Unknown font: %s", fname)
                return None

        family, pos = guessFont(fname)
        variants = [
            get_nt_fname(family) or fontfile,
            get_nt_fname(family + " Bold") or fontfile,
            get_nt_fname(family + " Italic")
            or get_nt_fname(family + " Oblique")
            or fontfile,
            get_nt_fname(family + " Bold Italic")
            or get_nt_fname(family + " Bold Oblique")
            or fontfile,
        ]
        return variants


def autoEmbed(fname):
    """Given a font name, does a best-effort of embedding
    said font and its variants.

    Returns a list of the font names it registered with ReportLab.

    """
    log.info("Trying to embed %s" % fname)
    fontList = []
    variants = []
    f = findFont(fname)
    if f:  # We have this font located
        if f[0].lower().endswith(".afm"):  # Type 1 font
            family = families[f[2]]

            # Register the whole family of faces
            faces = [
                pdfmetrics.EmbeddedType1Face(*fonts[fn.lower()][:2]) for fn in family
            ]
            for face in faces:
                pdfmetrics.registerTypeFace(face)

            for face, name in zip(faces, family):
                fontList.append(name)
                font = pdfmetrics.Font(face, name, "WinAnsiEncoding")
                log.info("Registering font: %s from %s" % (name, face.getFontFiles()))
                pdfmetrics.registerFont(font)

            # Map the variants
            regular, italic, bold, bolditalic = family
            for n in fname, regular:
                addMapping(n, 0, 0, regular)
                addMapping(n, 0, 1, italic)
                addMapping(n, 1, 0, bold)
                addMapping(n, 1, 1, bolditalic)
            log.info("Embedding as %s" % fontList)
            return fontList
        else:  # A TTF font
            variants = [fonts[x.lower()][0] for x in families[f[2]]]
    if not variants:  # Try fc-match
        variants = findTTFont(fname)
        if variants:
            vname = os.path.basename(variants[0])[:-4]
            if vname != fname:
                log.warning(
                    "Could not find font '%s'. Substituting with %s." % (fname, vname)
                )
    # It is a TT Font and we found it using fc-match (or found *something*)
    if variants:
        for variant in variants:
            vname = os.path.basename(variant)[:-4]
            try:
                if vname not in pdfmetrics._fonts:
                    _font = TTFont(vname, variant)
                    log.info("Registering font: %s from %s" % (vname, variant))
                    pdfmetrics.registerFont(_font)
            except TTFError:
                log.error("Error registering font: %s from %s" % (vname, variant))
            else:
                fontList.append(vname)
        regular, bold, italic, bolditalic = [
            os.path.basename(variant)[:-4] for variant in variants
        ]
        addMapping(regular, 0, 0, regular)
        addMapping(regular, 0, 1, italic)
        addMapping(regular, 1, 0, bold)
        addMapping(regular, 1, 1, bolditalic)
        log.info("Embedding via findTTFont as %s" % fontList)
    return fontList


def guessFont(fname):
    """Given a font name like "Tahoma-BoldOblique", "Bitstream Charter Italic"
    or "Perpetua Bold Italic" guess what it means.

    Returns (family, x) where x is
        0: regular
        1: bold
        2: italic
        3: bolditalic

    """
    italic = 0
    bold = 0
    if "-" not in fname:
        sfx = {
            "Bold": 1,
            "Bold Italic": 3,
            "Bold Oblique": 3,
            "Italic": 2,
            "Oblique": 2,
        }
        for key in sfx:
            if fname.endswith(" " + key):
                return fname.rpartition(key)[0], sfx[key]
        return fname, 0

    else:
        family, mod = fname.rsplit("-", 1)

    mod = mod.lower()
    if "oblique" in mod or "italic" in mod:
        italic = 1
    if "bold" in mod:
        bold = 1

    if bold + italic == 0:  # Not really a modifier
        return fname, 0
    return family, bold + 2 * italic


def main():
    global flist
    if len(sys.argv) != 2:
        print("Usage: findfonts.py fontName")
        sys.exit(1)
    if os.name == "nt":
        flist = [".", os.environ.get("SystemRoot", "C:\\Windows") + "\\Fonts"]
    else:
        flist = [".", "/usr/share/fonts", "/usr/share/texmf-dist/fonts"]
    fname = make_string(sys.argv[1])
    fn, _ = guessFont(fname)
    f = findFont(fn)
    if not f:
        f = findTTFont(fn)
    if f:
        print(f)
    else:
        print("Unknown font %s" % fname)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['rst2pdf.pdfbuilder']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'sphinxsubtitle'
copyright = u'2023, rst2pdf project'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '1.1'
# The full version, including alpha/beta/rc tags.
release = '1.0-test-release'


# -- Options for PDF output --------------------------------------------------

# Grouping the document tree into PDF files. List of tuples
# (source start file, target name, title, author, options).
#
# If there is more than one author, separate them with \\.
# For example: r'Guido van Rossum\\Fred L. Drake, Jr., editor'
#
# The options element is a dictionary that lets you override
# this config per-document. For example:
#
# ('index', 'MyProject', 'My Project', 'Author Name', {'pdf_compressed': True})
#
# would mean that specific document would be compressed
# regardless of the global 'pdf_compressed' setting.

pdf_documents = [
    ('index', 'MyProject', 'My Project', 'Author Name'),
]

# A comma-separated list of custom stylesheets. Example:
pdf_stylesheets = ['sphinx', 'a4']

# A list of folders to search for stylesheets. Example:
#pdf_style_path = ['.', '_styles']

# Create a compressed PDF
# Use True/False or 1/0
# Example: compressed=True
# pdf_compressed = False

# A colon-separated list of folders to search for fonts. Example:
# pdf_font_path = ['/usr/share/fonts', '/usr/share/texmf-dist/fonts/']

# Language to be used for hyphenation support
# pdf_language = "en_US"

# Mode for literal blocks wider than the frame. Can be
# overflow, shrink or truncate
# pdf_fit_mode = "shrink"

# Section level that forces a break page.
# For example: 1 means top-level sections start in a new page
# 0 means disabled
# pdf_break_level = 0

# When a section starts in a new page, force it to be 'even', 'odd',
# or just use 'any'
# pdf_breakside = 'any'

# Insert footnotes where they are defined instead of
# at the end.
# pdf_inline_footnotes = True

# verbosity level. 0 1 or 2
# pdf_verbosity = 0

# If false, no index is generated.
# pdf_use_index = True

# If false, no modindex is generated.
# pdf_use_modindex = True

# If false, no coverpage is generated.
# pdf_use_coverpage = True

# Name of the cover page template to use
# pdf_cover_template = 'sphinxcover.tmpl'

# Label to use as a prefix for the subtitle on the cover page
subtitle_prefix = 'Subtle Subtitle Prefix'

# Documents to append as an appendix to all manuals.
# pdf_appendices = []

# Enable experimental feature to split table cells. Use it
# if you get "DelayedTable too big" errors
# pdf_splittables = False

# Set the default DPI for images
# pdf_default_dpi = 72

# Enable rst2pdf extension modules
# pdf_extensions = []

# Page template name for "regular" pages
# pdf_page_template = 'cutePage'

# Show Table Of Contents at the beginning?
# pdf_use_toc = True

# How many levels deep should the table of contents be?
pdf_toc_depth = 9999

# Add section number to section references
pdf_use_numbered_links = False

# Background images fitting mode
pdf_fit_background_mode = 'scale'

# Repeat table header on tables that cross a page boundary?
pdf_repeat_table_rows = True

# Enable smart quotes (1, 2 or 3) or disable by setting to 0
pdf_smartquotes = 0

# Set a consistent date for the cover page
today = 'August 20, 2023'



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

import logging

log = logging.getLogger('rst2pdf')
_fmt = logging.Formatter('[%(levelname)s] %(filename)s:%(lineno)d %(message)s')
_hdlr = logging.StreamHandler()
_hdlr.setFormatter(_fmt)
log.addHandler(_hdlr)
log.setLevel(logging.WARNING)


def nodeid(node):
    """Given a node, tries to return a way to see where it was in the
    source text"""
    fname = 'UNKNOWN'
    line = 'UNKNOWN'

    try:
        if node.line:
            line = str(node.line)
    except Exception:
        pass

    try:
        if node.source:
            fname = str(node.source)
    except Exception:
        pass

    return 'near line %s in file %s' % (line, fname)



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

# Import all node handler modules here.
# The act of importing them wires them in.

from . import genelements  # noqa
from . import genpdftext  # noqa

# sphinxnodes needs these
from .genpdftext import NodeHandler, FontHandler, HandleEmphasis  # noqa

# createpdf needs this
nodehandlers = NodeHandler()



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

import os
import re
import tempfile

from reportlab.platypus.flowables import Flowable
from reportlab.platypus import SimpleDocTemplate
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont

from .log import log

try:
    from matplotlib import mathtext
    from matplotlib.font_manager import FontProperties
    from matplotlib.colors import to_rgb

    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False

fonts = {}


def enclose(s):
    """Enclose the string in $...$ if needed"""
    if not re.match(r'.*\$.+\$.*', s, re.MULTILINE | re.DOTALL):
        s = u"$%s$" % s
    return s


class Math(Flowable):
    def __init__(self, s, label=None, style=None):
        Flowable.__init__(self)
        self.s = s.strip()
        self.label = label
        self.fontsize = 10
        self.color = (0, 0, 0)
        self.hAlign = 'LEFT'
        if style:
            self.style = style
            self.fontsize = style.fontSize
            self.color = style.textColor.rgb()
            self.hAlign = style.alignment

        if HAS_MATPLOTLIB:
            self.parser = mathtext.MathTextParser("Path")
        else:
            log.error(
                "Math support not available,"
                " some parts of this document will be rendered incorrectly."
                " Install matplotlib."
            )

        if "\n" in s:
            log.error("rst2pdf's math directive does not support multiple lines'")

    def wrap(self, aW, aH):
        if HAS_MATPLOTLIB:
            try:
                (width, height, descent, _, _,) = self.parser.parse(
                    enclose(self.s), 72, prop=FontProperties(size=self.fontsize)
                )
                return width, height + descent
            except Exception as e:
                log.error(f"Math error in wrap: {e}")
                pass

        return 10, 10

    def drawOn(self, canv, x, y, _sW=0):
        if _sW and hasattr(self, 'hAlign'):
            from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY

            a = self.hAlign
            if a in ('CENTER', 'CENTRE', TA_CENTER):
                x = x + 0.5 * _sW
            elif a in ('RIGHT', TA_RIGHT):
                x = x + _sW
            elif a not in ('LEFT', TA_LEFT, TA_JUSTIFY):
                raise ValueError("Bad hAlign value " + str(a))
        height = 0
        if HAS_MATPLOTLIB:
            global fonts
            canv.saveState()
            try:
                (width, height, descent, glyphs, rects,) = self.parser.parse(
                    enclose(self.s), 72, prop=FontProperties(size=self.fontsize)
                )
                canv.translate(x, y + descent)

                for font, fontsize, num, ox, oy in glyphs:
                    fontname = font.fname
                    if fontname not in fonts:
                        fonts[fontname] = fontname
                        pdfmetrics.registerFont(TTFont(fontname, fontname))
                    canv.setFont(fontname, fontsize)
                    rgb_color = to_rgb(self.color)
                    canv.setFillColorRGB(rgb_color[0], rgb_color[1], rgb_color[2])
                    canv.drawString(ox, oy, chr(num))

                canv.setLineWidth(0)
                canv.setDash([])
                for ox, oy, width, height in rects:
                    canv.rect(ox, oy + height, width, height, fill=1)
            except Exception as e:
                log.error(f"Math error: {e}")
                log.exception("Math error!")
                canv.translate(x, y)
                rgb_color = to_rgb(self.color)
                canv.setFillColorRGB(rgb_color[0], rgb_color[1], rgb_color[2])
                canv.drawString(0, 0, self.s)
            canv.restoreState()
        else:
            canv.saveState()
            canv.translate(x, y)
            canv.drawString(x, y, self.s)
            canv.restoreState()
        if self.label:
            log.info('Drawing equation-%s' % self.label)
            canv.bookmarkHorizontal('equation-%s' % self.label, 0, height)

    def descent(self):
        """Return the descent of this flowable,
        useful to align it when used inline."""
        if HAS_MATPLOTLIB:
            width, height, descent, glyphs, rects = self.parser.parse(
                enclose(self.s), 72, prop=FontProperties(size=self.fontsize)
            )
            return descent
        return 0

    def genImage(self):
        """Create a PNG from the contents of this flowable.

        Required so we can put inline math in paragraphs.
        Returns the file name.
        The file is caller's responsibility.

        """

        dpi = 72
        scale = 72

        try:
            import Image
            import ImageFont
            import ImageDraw
        except ImportError:
            from PIL import (
                Image,
                ImageFont,
                ImageDraw,
            )

        if not HAS_MATPLOTLIB:
            img = Image.new('RGBA', (120, 120), (255, 255, 255, 0))
        else:
            (width, height, descent, glyphs, rects,) = self.parser.parse(
                enclose(self.s), dpi, prop=FontProperties(size=self.fontsize)
            )
            img = Image.new(
                'RGBA',
                (int(width * scale), int((height + descent) * scale)),
                (255, 255, 255, 0),
            )
            draw = ImageDraw.Draw(img)
            for font, fontsize, num, ox, oy in glyphs:
                fontname = font.fname
                image_font = ImageFont.truetype(fontname, int(fontsize * scale))
                fc = to_rgb(self.color)
                rgb_color = (int(fc[0] * 255), int(fc[1] * 255), int(fc[2] * 255))
                draw.text(
                    (ox * scale, (height + 1 - oy - fontsize) * scale),
                    chr(num),
                    font=image_font,
                    fill=rgb_color,
                )
            for ox, oy, w, h in rects:
                x1 = ox * scale
                x2 = x1 + w * scale
                y1 = (height - 1 - oy) * scale
                y2 = y1 + (h * scale)
                draw.rectangle([x1, y1, x2, y2], (0, 0, 0))

        fh, fn = tempfile.mkstemp(suffix=".png")
        os.close(fh)
        img.save(fn)
        return fn


if __name__ == "__main__":
    doc = SimpleDocTemplate("mathtest.pdf")
    Story = [
        Math(r'\mathcal{R}\prod_{i=\alpha\mathcal{B}}' r'^\infty a_i\sin(2 \pi f x_i)')
    ]
    doc.build(Story)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Run all tests under a profiling environment."""

import cProfile
import os

from rst2pdf.createpdf import RstToPdf


def run():
    inpdir = os.path.abspath('./input')
    outdir = os.path.abspath('./tmp')
    # Discard output, this is not about whether things
    # work or not, that's testing ;-)

    for f in os.listdir(inpdir):
        if f.endswith('.rst'):  # Test case
            print('Running: %s' % f)
            sheet = os.path.join(inpdir, f[:-4] + '.style')
            if os.path.exists(sheet):
                sheet = [sheet]
            else:
                sheet = []

            r2p = RstToPdf(stylesheets=sheet)
            try:
                fname = os.path.join(inpdir, f)
                r2p.createPdf(
                    text=open(fname).read(),
                    output=os.path.join(outdir, f + '.pdf'),
                    source_path=fname,
                )
            except Exception:
                print('FAIL')


cProfile.runctx("run()", globals(), locals(), filename="rst2pdf.profile")



# -*- coding: utf-8 -*-

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.todo',
    'sphinx.ext.imgmath',
    'sphinx.ext.graphviz',
    'rst2pdf.pdfbuilder',
]

# The master toctree document.
master_doc = 'foobar'

# General information about the project.
project = u'Foobar'
copyright = u'2009, Jason S'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '1.0.1'
# The full version, including alpha/beta/rc tags.
release = '1.0.1'


# -- Options for PDF output ----------------------------------------------------

# Grouping the document tree into PDF files. List of tuples
# (source start file, target name, title, author, options).
pdf_documents = [('foobar', u'foobar', u'foobar Documentation', u'jsachs')]

# Language to be used for hyphenation support
pdf_language = "en_US"

# verbosity level. 0 1 or 2
pdf_verbosity = 0

pdf_invariant = True
pdf_real_footnotes = True

# Set a consistent date for the cover page
today = 'April 29, 2018'


# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.platypus import Flowable, Paragraph
from svglib.svglib import svg2rlg


class SVGImage(Flowable):
    def __init__(
        self,
        filename,
        width=None,
        height=None,
        kind='direct',
        mask=None,
        lazy=True,
        srcinfo=None,
    ):
        Flowable.__init__(self)
        self._kind = kind
        self._mode = 'svg2rlg'
        self.doc = svg2rlg(filename)
        self.imageWidth = width
        self.imageHeight = height
        x1, y1, x2, y2 = self.doc.getBounds()
        # Actually, svg2rlg's getBounds seems broken.
        self._w, self._h = x2, y2
        if not self.imageWidth:
            self.imageWidth = self._w
        if not self.imageHeight:
            self.imageHeight = self._h
        self.__ratio = float(self.imageWidth) / self.imageHeight
        if kind in ['direct', 'absolute']:
            self.drawWidth = width or self.imageWidth
            self.drawHeight = height or self.imageHeight
        elif kind in ['bound', 'proportional']:
            factor = min(
                float(width) / self.imageWidth,
                float(height) / self.imageHeight,
            )
            self.drawWidth = self.imageWidth * factor
            self.drawHeight = self.imageHeight * factor

    def wrap(self, aW, aH):
        return self.drawWidth, self.drawHeight

    def drawOn(self, canv, x, y, _sW=0):
        if _sW and hasattr(self, 'hAlign'):
            a = self.hAlign
            if a in ('CENTER', 'CENTRE', TA_CENTER):
                x += 0.5 * _sW
            elif a in ('RIGHT', TA_RIGHT):
                x += _sW
            elif a not in ('LEFT', TA_LEFT):
                raise ValueError("Bad hAlign value " + str(a))
        canv.saveState()
        canv.translate(x, y)
        canv.scale(self.drawWidth / self._w, self.drawHeight / self._h)
        self.doc._drawOn(canv)
        canv.restoreState()


if __name__ == "__main__":
    import sys
    from reportlab.platypus import SimpleDocTemplate
    from reportlab.lib.styles import getSampleStyleSheet

    doc = SimpleDocTemplate('svgtest.pdf')
    styles = getSampleStyleSheet()
    style = styles['Normal']
    Story = [
        Paragraph("Before the image", style),
        SVGImage(sys.argv[1]),
        Paragraph("After the image", style),
    ]
    doc.build(Story)



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

'''
This module provides one useful class:  NodeHandler

The NodeHandler class is designed to be subclassed.  Each subclass
should support the processing that createpdf.RstToPdf needs to do
on a particular type of node that could appear in a document tree.

When the subclass is defined, it should reference NodeHandler as
the first base class, and one or more docutils node classes as
subsequent base classes.

These docutils node classes will not actually wind up in the
base classes of the subclass.  Instead, they will be used as
keys in a dispatch dictionary which is used to find the correct
NodeHandler subclass to use to process an instance of a given
docutils node class.

When an instance of createpdf.RstToPdf is created, a NodeHandler
instance will be called to return dispatchers for gather_elements
and gather_pdftext, wrapped up as methods of the createpdf.RstToPdf
class.

When a dispatcher is called, it will dispatch to the correct subclass
to handle the given docutils node instance.

If no NodeHandler subclass has been created to handle that particular
type of docutils node, then default processing will occur and a warning
will be logged.
'''

from copy import copy
import inspect
import types

import docutils.nodes
from docutils.utils import smartquotes

from .flowables import BoundByWidth, TocEntry
from .log import log, nodeid


class MetaHelper(type):
    """MetaHelper is designed to generically enable a few of the benefits of
    using metaclasses by encapsulating some of the complexity of setting
    them up.

    If a base class uses MetaHelper (by assigning __metaclass__ = MetaHelper),
    then that class (and its metaclass inheriting subclasses) can control
    class creation behavior by defining a couple of helper functions.

    1) A base class can define a _classpreinit function.  This function
       is called during __new__ processing of the class object itself,
       but only during subclass creation (not when the class defining
       the _classpreinit is itself created).

       The subclass object does not yet exist at the time _classpreinit
       is called.  _classpreinit accepts all the parameters of the
       __new__ function for the class itself (not the same as the __new__
       function for the instantiation of class objects!) and must return
       a tuple of the same objects.  A typical use of this would be to
       modify the class bases before class creation.

    2) Either a base class or a subclass can define a _classinit() function.
       This function will be called immediately after the actual class has
       been created, and can do whatever setup is required for the class.
       Note that every base class (but not every subclass) which uses
       MetaHelper MUST define _classinit, even if that definition is None.

     MetaHelper also places an attribute into each class created with it.
     _baseclass is set to None if this class has no superclasses which
     also use MetaHelper, or to the first such MetaHelper-using baseclass.
     _baseclass can be explicitly set inside the class definition, in
     which case MetaHelper will not override it.
    """

    def __new__(clstype, name, bases, clsdict):
        # Our base class is the first base in the class definition which
        # uses MetaHelper, or None if no such base exists.
        base = ([x for x in bases if type(x) is MetaHelper] + [None])[0]

        # Only set our base into the class if it has not been explicitly
        # set
        clsdict.setdefault('_baseclass', base)

        # See if the base class definied a preinit function, and call it
        # if so.
        preinit = getattr(base, '_classpreinit', None)
        if preinit is not None:
            clstype, name, bases, clsdict = preinit(clstype, name, bases, clsdict)

        # Delegate the real work to type
        return type.__new__(clstype, name, bases, clsdict)

    def __init__(cls, name, bases, clsdict):
        # Let type build the class for us
        type.__init__(cls, name, bases, clsdict)
        # Call the class's initialization function if defined
        if cls._classinit is not None:
            cls._classinit()


class NodeHandler(metaclass=MetaHelper):
    """NodeHandler classes are used to dispatch
    to the correct class to handle some node class
    type, via a dispatchdict in the main class.
    """

    dispatchdict = {}

    @classmethod
    def _classpreinit(baseclass, clstype, name, bases, clsdict):
        # _classpreinit is called before the actual class is built
        # Perform triage on the class bases to separate actual
        # inheritable bases from the target docutils node classes
        # which we want to dispatch for.

        new_bases = []
        targets = []
        for target in bases:
            if target is not object:
                (targets, new_bases)[issubclass(target, NodeHandler)].append(target)
        clsdict['_targets'] = targets
        return clstype, name, tuple(new_bases), clsdict

    @classmethod
    def _classinit(cls):
        # _classinit() is called once the subclass has actually
        # been created.

        # For the base class, just add a dispatch dictionary
        if cls._baseclass is None:
            cls.dispatchdict = {}
            return

        # for subclasses, instantiate them, and then add
        # the class to the dispatch dictionary for each of its targets.
        self = cls()
        for target in cls._targets:
            if cls.dispatchdict.setdefault(target, self) is not self:
                t = repr(target)
                old = repr(cls.dispatchdict[target])
                new = repr(self)
                log.debug(
                    'Dispatch handler %s for node type %s overridden by %s'
                    % (old, t, new)
                )
                cls.dispatchdict[target] = self

    @staticmethod
    def getclassname(obj):
        cln = repr(obj.__class__)
        info = cln.split("'")
        if len(info) == 3:
            return info[1]
        return cln

    def log_unknown(self, node, during):
        if not hasattr(self, 'unkn_node'):
            self.unkn_node = set()
        cln = self.getclassname(node)
        if cln not in self.unkn_node:
            self.unkn_node.add(cln)
            log.warning("Unkn. node (self.%s): %s [%s]", during, cln, nodeid(node))
            try:
                log.debug(node)
            except (UnicodeDecodeError, UnicodeEncodeError):
                log.debug(repr(node))

    def findsubclass(self, node, during):
        handlerinfo = '%s.%s' % (self.getclassname(self), during)
        log.debug("%s: %s", handlerinfo, self.getclassname(node))
        log.debug("%s: [%s]", handlerinfo, nodeid(node))
        try:
            log.debug("%s: %s", handlerinfo, node)
        except (UnicodeDecodeError, UnicodeEncodeError):
            log.debug("%s: %r", handlerinfo, node)
        log.debug("")

        # Dispatch to the first matching class in the MRO

        dispatchdict = self.dispatchdict
        for baseclass in inspect.getmro(node.__class__):
            result = dispatchdict.get(baseclass)
            if result is not None:
                break
        else:
            self.log_unknown(node, during)
            result = self
        return result

    def __call__(self, client):
        '''Get the dispatchers, wrapped up as methods for the client'''
        textdispatch = types.MethodType(self.textdispatch, client)
        elemdispatch = types.MethodType(self.elemdispatch, client)
        return textdispatch, elemdispatch

    # This overridable attribute will be set true in the instance
    # if handling a sphinx document

    sphinxmode = False

    # Begin overridable attributes and methods for elemdispatch

    def gather_elements(self, client, node, style):
        return client.gather_elements(node, style=style)

    def _get_non_default_values_from_style(self, style):
        """Return a dictionary of all the items in the style that are changed from their default value"""
        items = style.__dict__.copy()
        defaults = style.defaults
        for key in defaults.keys():
            if defaults[key] == items[key]:
                del items[key]
        return items

    def getstyle(self, client, node, style):
        try:
            if node['classes'] and node['classes'][0]:
                for n in range(len(node['classes'])):
                    if node['classes'][n] in client.styles.StyleSheet:
                        if n == 0:
                            style = client.styles[node['classes'][n]]
                        else:
                            # merge the non-default properties of this style into the style we currently have
                            style = copy(style)
                            items = self._get_non_default_values_from_style(
                                client.styles[node['classes'][n]]
                            )
                            items.pop("parent", None)
                            name = f"{style.__dict__['name']}-{items['name']}"
                            items['name'] = name
                            style.__dict__.update(items)
                    else:
                        log.info(
                            "Unknown class %s, ignoring. [%s]",
                            node['classes'][n],
                            nodeid(node),
                        )
        except TypeError:  # Happens when a docutils.node.Text reaches here
            pass

        if style is None or style == client.styles['bodytext']:
            style = client.styles.styleForNode(node)
        return style

    def getelements(self, client, node, style):
        style = self.getstyle(client, node, style)
        elements = self.gather_elements(client, node, style)

        # Make all the sidebar cruft unreachable
        # if style.__dict__.get('float','None').lower() !='none':
        # node.elements=[Sidebar(node.elements,style)]
        # elif 'width' in style.__dict__:

        if 'width' in style.__dict__:
            elements = [BoundByWidth(style.width, elements, style, mode="shrink")]

        return elements

    # End overridable attributes and methods for elemdispatch

    def elemdispatch(self, client, node, style=None):
        self = self.findsubclass(node, 'elemdispatch')

        # set anchors for internal references
        try:
            for i in node['ids']:
                client.pending_targets.append(i)
        except TypeError:  # Happens with docutils.node.Text
            pass

        elements = self.getelements(client, node, style)

        if node.line and client.debugLinesPdf:
            elements.insert(0, TocEntry(client.depth - 1, 'LINE-%s' % node.line))
        node.elements = elements
        return elements

    # Begin overridable attributes and methods for textdispatch

    pre = ''
    post = ''

    def get_pre_post(self, client, node, replaceEnt):
        return self.pre, self.post

    def get_text(self, client, node, replaceEnt):
        return client.gather_pdftext(node)

    def apply_replacements(self, text, smarty, node):
        # Try to be clever about when to replace characters
        if node.__class__ in (
            docutils.nodes.paragraph,
            docutils.nodes.block_quote,
            docutils.nodes.title,
        ):
            return smartquotes.smartyPants(text, smarty)
        return text

    # End overridable attributes and methods for textdispatch

    def textdispatch(self, client, node, replaceEnt=True):
        self = self.findsubclass(node, 'textdispatch')
        pre, post = self.get_pre_post(client, node, replaceEnt)
        text = self.get_text(client, node, replaceEnt)
        text = pre + text + post

        try:
            log.debug("%s.textdispatch: %s" % (self.getclassname(self), text))
        except UnicodeDecodeError:
            pass

        text = self.apply_replacements(text, client.smartypants_attributes, node)
        node.pdftext = text
        return text



# -*- coding: utf-8 -*-
# Copyright (c) 2021 by Rob Allen
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.parts import Contents as BaseContents


class Contents(BaseContents):
    """Override the standard docutils contents directive to get access to the depth option

    If it is not set, then assign to a very large number which will effectively be unlimited.
    """

    depth = 9999

    def __init__(
        self,
        name,
        arguments,
        options,
        content,
        lineno,
        content_offset,
        block_text,
        state,
        state_machine,
    ):
        super().__init__(
            name,
            arguments,
            options,
            content,
            lineno,
            content_offset,
            block_text,
            state,
            state_machine,
        )
        Contents.depth = options['depth'] if 'depth' in options else Contents.depth


directives._directives['contents'] = Contents



# -*- coding: utf-8 -*-

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['rst2pdf.pdfbuilder']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'sample'
copyright = u'2010, sample'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '1.0'
# The full version, including alpha/beta/rc tags.
release = '1.0'


# -- Options for PDF output ----------------------------------------------------

# Grouping the document tree into PDF files. List of tuples
# (source start file, target name, title, author).
pdf_documents = [('index', u'sample', u'sample Documentation', u'sample')]
pdf_stylesheets = ['sphinx', 'a4', 'ja']

pdf_font_path = ['../..']

pdf_invariant = True
pdf_real_footnotes = True

# Set a consistent date for the cover page
today = 'April 29, 2018'


# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

"""Singleton config object"""

import configparser
import os

from rst2pdf.rson import loads


cfdir = os.path.join(os.path.expanduser('~'), '.rst2pdf')
cfname = os.path.join(cfdir, 'config')


def getValue(section, key, default=None):
    section = section.lower()
    key = key.lower()
    try:
        return loads(conf.get(section, key))
    except Exception:
        return default


class ConfigError(Exception):
    def __init__(self, modulename, msg):
        self.modulename = modulename
        self.msg = msg


conf = configparser.ConfigParser()


def parseConfig(extracf=None):
    global conf
    cflist = ["/etc/rst2pdf.conf", cfname]
    if extracf:
        cflist.append(extracf)
    conf = configparser.ConfigParser()
    conf.read(cflist)


parseConfig()



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

import shlex

import jinja2
from reportlab.lib.colors import Color
from reportlab.platypus.flowables import CondPageBreak

from . import flowables
from .log import log, nodeid
from .styles import adjustUnits

PageCounter = None


def parseRaw(data, node):
    """Parse and process a simple DSL to handle creation of flowables.

    Supported (can add others on request):

    * PageBreak [<template name>] [background={img_filename.jpg}] [background_fit_mode={mode (e.g. stretch)}]
    * EvenPageBreak
    * OddPageBreak
    * FrameBreak
    * Spacer width, height
    * Transition
    * SetPageCounter
    * TextAnnotation "text of annotation" x_begin=-1 y_begin=-1 x_end=-1 y_end=-1

    """
    global PageCounter
    if PageCounter is None:
        from rst2pdf.createpdf import PageCounter as pc

        PageCounter = pc

    elements = []
    lines = data.splitlines()
    for line in lines:
        lexer = shlex.shlex(line, punctuation_chars=True)
        lexer.whitespace += ','
        tokens = list(lexer)
        if not tokens:
            continue  # Empty line
        command = tokens[0]
        if command == 'PageBreak':
            if len(tokens) == 1:
                elements.append(flowables.MyPageBreak())
            else:
                elements.append(flowables.MyPageBreak(tokens[1:]))
        elif command == 'EvenPageBreak':
            if len(tokens) == 1:
                elements.append(flowables.MyPageBreak(breakTo='even'))
            else:
                elements.append(flowables.MyPageBreak(tokens[1:], breakTo='even'))
        elif command == 'OddPageBreak':
            if len(tokens) == 1:
                elements.append(flowables.MyPageBreak(breakTo='odd'))
            else:
                elements.append(flowables.MyPageBreak(tokens[1:], breakTo='odd'))
        elif command == 'FrameBreak':
            if len(tokens) == 1:
                elements.append(CondPageBreak(99999))
            else:
                elements.append(CondPageBreak(float(tokens[1])))
        elif command == 'Spacer':
            elements.append(
                flowables.MySpacer(adjustUnits(tokens[1]), adjustUnits(tokens[2]))
            )
        elif command == 'Transition':
            elements.append(flowables.Transition(*tokens[1:]))
        elif command == 'SetPageCounter':
            elements.append(PageCounter(*tokens[1:]))
        elif command == 'TextAnnotation':
            elements.append(flowables.TextAnnotation(*tokens[1:]))
        else:
            log.error(
                'Unknown command %s in raw pdf directive [%s]' % (command, nodeid(node))
            )
    return elements


HAS_XHTML2PDF = True
try:
    import xhtml2pdf.default  # NOQA
    from xhtml2pdf.util import COLOR_BY_NAME
    from xhtml2pdf.util import Memoized as memoized
    from xhtml2pdf.context import pisaContext
    from xhtml2pdf.parser import pisaGetAttributes
    from xhtml2pdf.document import pisaStory
    import xhtml2pdf.parser as pisa_parser
except ImportError:
    try:
        from sx.pisa3.pisa_util import COLOR_BY_NAME

        def memoized(*a):
            return a

        from sx.pisa3.pisa_context import pisaContext
        from sx.pisa3.pisa_parser import pisaGetAttributes
        from sx.pisa3.pisa_document import pisaStory
        import sx.pisa3.pisa_parser as pisa_parser
    except ImportError:
        HAS_XHTML2PDF = False


if HAS_XHTML2PDF:

    COLOR_BY_NAME['initial'] = Color(0, 0, 0)

    from xml.dom import Node

    def pisaPreLoop2(node, context, collect=False):
        """
        Collect all CSS definitions
        """

        data = u""
        if node.nodeType == Node.TEXT_NODE and collect:
            data = node.data

        elif node.nodeType == Node.ELEMENT_NODE:
            name = node.tagName.lower()

            # print name, node.attributes.items()
            if name in ("style", "link"):
                attr = pisaGetAttributes(context, name, node.attributes)
                print(" ", attr)
                media = [x.strip() for x in attr.media.lower().split(",") if x.strip()]
                # print repr(media)

                if attr.get("type", "").lower() in ("", "text/css") and (
                    not media or "all" in media or "print" in media or "pdf" in media
                ):

                    if name == "style":
                        for node in node.childNodes:
                            data += pisaPreLoop2(node, context, collect=True)
                        return u""

                    if (
                        name == "link"
                        and attr.href
                        and attr.rel.lower() == "stylesheet"
                    ):
                        context.addCSS(
                            '\n@import "%s" %s;' % (attr.href, ",".join(media))
                        )

        for node in node.childNodes:
            result = pisaPreLoop2(node, context, collect=collect)
            if collect:
                data += result

        return data

    pisa_parser.pisaPreLoop = pisaPreLoop2

    HTML_CSS = """
    html {
        font-family: Helvetica;
        font-size: 7px;
        font-weight: normal;
        color: #000000;
        background-color: transparent;
        margin: 0;
        padding: 0;
        line-height: 150%;
        border: 1px none;
        display: inline;
        width: auto;
        height: auto;
        white-space: normal;
    }

    b,
    strong {
        font-weight: bold;
    }

    i,
    em {
        font-style: italic;
    }

    u {
        text-decoration: underline;
    }

    s,
    strike {
        text-decoration: line-through;
    }

    a {
        text-decoration: underline;
        color: blue;
    }

    ins {
        color: green;
        text-decoration: underline;
    }
    del {
        color: red;
        text-decoration: line-through;
    }

    pre,
    code,
    kbd,
    samp,
    tt {
        font-family: "Courier New";
    }

    h1,
    h2,
    h3,
    h4,
    h5,
    h6 {
        font-weight:bold;
        -pdf-outline: true;
        -pdf-outline-open: false;
    }

    h1 {
        /*18px via YUI Fonts CSS foundation*/
        font-size:138.5%;
        -pdf-outline-level: 0;
    }

    h2 {
        /*16px via YUI Fonts CSS foundation*/
        font-size:123.1%;
        -pdf-outline-level: 1;
    }

    h3 {
        /*14px via YUI Fonts CSS foundation*/
        font-size:108%;
        -pdf-outline-level: 2;
    }

    h4 {
        -pdf-outline-level: 3;
    }

    h5 {
        -pdf-outline-level: 4;
    }

    h6 {
        -pdf-outline-level: 5;
    }

    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    p,
    pre,
    hr {
        margin:1em 0;
    }

    address,
    blockquote,
    body,
    center,
    dl,
    dir,
    div,
    fieldset,
    form,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    hr,
    isindex,
    menu,
    noframes,
    noscript,
    ol,
    p,
    pre,
    table,
    th,
    tr,
    td,
    ul,
    li,
    dd,
    dt,
    pdftoc {
        display: block;
    }

    table {
    }

    tr,
    th,
    td {

        vertical-align: middle;
        width: auto;
    }

    th {
        text-align: center;
        font-weight: bold;
    }

    center {
        text-align: center;
    }

    big {
        font-size: 125%;
    }

    small {
        font-size: 75%;
    }


    ul {
        margin-left: 1.5em;
        list-style-type: disc;
    }

    ul ul {
        list-style-type: circle;
    }

    ul ul ul {
        list-style-type: square;
    }

    ol {
        list-style-type: decimal;
        margin-left: 1.5em;
    }

    pre {
        white-space: pre;
    }

    blockquote {
        margin-left: 1.5em;
        margin-right: 1.5em;
    }

    noscript {
        display: none;
    }
    """

    def parseHTML(data, node):
        path = None
        link_callback = None
        debug = 0
        default_css = HTML_CSS
        xhtml = False
        encoding = None
        xml_output = None
        capacity = 100 * 1024

        # Prepare simple context
        context = pisaContext(path, debug=debug, capacity=capacity)
        context.pathCallback = link_callback

        # Build story
        context = pisaStory(
            data,
            path,
            link_callback,
            debug,
            default_css,
            xhtml,
            encoding,
            context=context,
            xml_output=xml_output,
        )
        return context.story

else:  # no xhtml2pdf

    def parseHTML(data, none):
        log.error("You need xhtml2pdf installed to use the raw HTML directive.")
        return []


class DependencyRecordingFileSystemLoader(jinja2.FileSystemLoader):
    def __init__(self, *args, record_dependencies=None, **kwargs):
        self.record_dependencies = record_dependencies
        super().__init__(*args, **kwargs)

    def get_source(self, environment, template):
        r = (_, path, _) = super().get_source(environment, template)
        if self.record_dependencies is not None:
            self.record_dependencies.add(path)
        return r



# -*- coding: utf-8 -*-
"""
      Sphinx rst2pdf builder extension
      ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      Usage:
      1. In conf.py add 'rst2pdf.pdfbuilder' element to 'extensions' list:
         extensions = ['rst2pdf.pdfbuilder']
      2. Modify your Makefile or run it with:
         $ sphinx-build -d_build/doctrees -bpdf . _build/pdf

    :copyright: Copyright 2009 Roberto Alsina, Wojtek Walczak
    :license: BSD, see LICENSE for details.
"""

from copy import copy
from io import BytesIO
import logging
import os
import os.path
import re
import sys
import time
from urllib.parse import urlunparse

from docutils import writers
from docutils import nodes
from docutils.transforms.parts import Contents
from docutils.io import FileOutput
import docutils.core
import jinja2
from pygments.lexers import guess_lexer
import sphinx
from sphinx import addnodes
from sphinx.builders import Builder
from sphinx.environment.adapters.indexentries import IndexEntries
from sphinx.locale import _
from sphinx.transforms import SphinxTransform
from sphinx.util.console import darkgreen, red  # type: ignore
from sphinx.util import SEP

import rst2pdf
from rst2pdf import createpdf
from rst2pdf.directives import code_block
from rst2pdf.log import log
from rst2pdf.languages import get_language_available

if sphinx.version_info >= (2, 1):
    from sphinx.errors import NoUri
else:
    from sphinx.environment import NoUri


class PDFBuilder(Builder):
    name = 'pdf'
    out_suffix = '.pdf'
    supported_image_types = ['image/svg+xml', 'image/png', 'image/gif', 'image/jpeg']

    def init(self):
        self.docnames = []
        self.document_data = []
        self.sphinx_logger = sphinx.util.logging.getLogger(__name__)

    def write(self, *ignored):

        self.init_document_data()

        if self.config.pdf_verbosity > 1:
            log.setLevel(logging.DEBUG)
        elif self.config.pdf_verbosity > 0:
            log.setLevel(logging.INFO)

        for entry in self.document_data:
            try:
                docname, targetname, title, author = entry[:4]
                # Custom options per document
                if len(entry) > 4 and isinstance(entry[4], dict):
                    opts = entry[4]
                else:
                    opts = {}
                self.sphinx_logger.info("processing " + targetname + "... ")
                self.opts = opts

                class dummy:
                    extensions = self.config.pdf_extensions

                createpdf.add_extensions(dummy())

                self.page_template = opts.get(
                    'pdf_page_template', self.config.pdf_page_template
                )

                docwriter = PDFWriter(
                    self,
                    stylesheets=opts.get(
                        'pdf_stylesheets', self.config.pdf_stylesheets
                    ),
                    language=opts.get('pdf_language', self.config.pdf_language),
                    breaklevel=opts.get('pdf_break_level', self.config.pdf_break_level),
                    breakside=opts.get('pdf_breakside', self.config.pdf_breakside),
                    fontpath=opts.get('pdf_font_path', self.config.pdf_font_path),
                    fitmode=opts.get('pdf_fit_mode', self.config.pdf_fit_mode),
                    compressed=opts.get('pdf_compressed', self.config.pdf_compressed),
                    inline_footnotes=opts.get(
                        'pdf_inline_footnotes', self.config.pdf_inline_footnotes
                    ),
                    splittables=opts.get(
                        'pdf_splittables', self.config.pdf_splittables
                    ),
                    repeat_table_rows=opts.get(
                        'pdf_repeat_table_rows', self.config.pdf_repeat_table_rows
                    ),
                    default_dpi=opts.get(
                        'pdf_default_dpi', self.config.pdf_default_dpi
                    ),
                    page_template=self.page_template,
                    invariant=opts.get('pdf_invariant', self.config.pdf_invariant),
                    real_footnotes=opts.get(
                        'pdf_real_footnotes', self.config.pdf_real_footnotes
                    ),
                    use_toc=opts.get('pdf_use_toc', self.config.pdf_use_toc),
                    toc_depth=opts.get('pdf_toc_depth', self.config.pdf_toc_depth),
                    use_coverpage=opts.get(
                        'pdf_use_coverpage', self.config.pdf_use_coverpage
                    ),
                    use_numbered_links=opts.get(
                        'pdf_use_numbered_links', self.config.pdf_use_numbered_links
                    ),
                    fit_background_mode=opts.get(
                        'pdf_fit_background_mode', self.config.pdf_fit_background_mode
                    ),
                    baseurl=opts.get('pdf_baseurl', self.config.pdf_baseurl),
                    section_header_depth=opts.get(
                        'section_header_depth', self.config.section_header_depth
                    ),
                    smartquotes=str(
                        opts.get('pdf_smartquotes', self.config.pdf_smartquotes)
                    ),
                    srcdir=self.srcdir,
                    style_path=opts.get('pdf_style_path', self.config.pdf_style_path),
                    config=self.config,
                )

                tgt_file = os.path.join(self.outdir, targetname + self.out_suffix)
                destination = FileOutput(
                    destination=open(tgt_file, 'wb'), encoding='utf-8'
                )
                doctree = self.assemble_doctree(
                    docname,
                    title,
                    author,
                    appendices=opts.get('pdf_appendices', self.config.pdf_appendices)
                    or [],
                )
                doctree.settings.author = author
                doctree.settings.title = title
                self.sphinx_logger.info("done")
                self.sphinx_logger.info("writing " + targetname + "... ")
                docwriter.write(doctree, destination)
                self.sphinx_logger.info("done")
            except Exception:
                log.exception('Failed to build doc')
                self.sphinx_logger.info(red("FAILED"))

    def init_document_data(self):
        preliminary_document_data = map(list, self.config.pdf_documents)
        if not preliminary_document_data:
            self.sphinx_logger.warning(
                'no "pdf_documents" config value found; no documents ' 'will be written'
            )
            return
        # assign subdirs to titles
        self.titles = []
        for entry in preliminary_document_data:
            docname = entry[0]
            if docname not in self.env.all_docs:
                self.sphinx_logger.warning(
                    '"pdf_documents" config value references unknown '
                    'document %s' % docname
                )
                continue
            self.document_data.append(entry)
            if docname.endswith(SEP + 'index'):
                docname = docname[:-5]
            self.titles.append((docname, entry[2]))

    def assemble_doctree(self, docname, title, author, appendices):

        # FIXME: use the new inline_all_trees from Sphinx.
        # check how the LaTeX builder does it.

        self.docnames = set([docname])
        self.sphinx_logger.info(darkgreen(docname) + " ")

        def process_tree(docname, tree):
            tree = tree.deepcopy()
            for toctreenode in tree.traverse(addnodes.toctree):
                newnodes = []
                includefiles = map(str, toctreenode['includefiles'])
                for includefile in includefiles:
                    try:
                        self.sphinx_logger.info(darkgreen(includefile) + " ")
                        subtree = process_tree(
                            includefile, self.env.get_doctree(includefile)
                        )
                        self.docnames.add(includefile)
                    except Exception:
                        self.sphinx_logger.warning(
                            '%s: toctree contains ref to nonexisting file %r'
                            % (docname, includefile)
                        )
                    else:
                        sof = addnodes.start_of_file(docname=includefile)
                        sof.children = subtree.children
                        newnodes.append(sof)
                toctreenode.parent.replace(toctreenode, newnodes)
            return tree

        tree = self.env.get_doctree(docname)
        tree = process_tree(docname, tree)

        self.docutils_languages = {}
        if self.config.language:
            self.docutils_languages[self.config.language] = get_language_available(
                self.config.language
            )[2]

        if self.opts.get('pdf_use_index', self.config.pdf_use_index):
            # Add index at the end of the document

            # This is a hack. create_index creates an index from
            # ALL the documents data, not just this one.
            # So, we preserve a copy, use just what we need, then
            # restore it.
            if hasattr(self.env, 'domains') and "index" in self.env.domains:
                # Sphinx 2.4.0+ stores the index entries in self.env.domains["index"]
                t = copy(self.env.domains["index"].data['entries'])
                try:
                    self.env.domains["index"].data["entries"] = {
                        docname: self.env.domains["index"].data['entries'][
                            docname + '-gen'
                        ]
                    }
                except KeyError:
                    self.env.domains["index"].data["entries"] = {}
                    for dname in self.docnames:
                        self.env.domains["index"].data["entries"][dname] = t.get(
                            dname, []
                        )

                genindex = IndexEntries(self.env).create_index(self)
                self.env.domains["index"].data['entries'] = t
            elif hasattr(self.env, 'indexentries'):
                # Sphinx 2.3.1 or lower stores the index entries in self.env.indexentries
                t = copy(self.env.indexentries)
                try:
                    self.env.indexentries = {
                        docname: self.env.indexentries[docname + '-gen']
                    }
                except KeyError:
                    self.env.indexentries = {}
                    for dname in self.docnames:
                        self.env.indexentries[dname] = t.get(dname, [])

                genindex = IndexEntries(self.env).create_index(self)
                self.env.indexentries = t
            # EOH (End Of Hack)

            if genindex:  # No point in creating empty indexes
                index_nodes = genindex_nodes(genindex)
                tree.append(nodes.raw(text='OddPageBreak twoColumn', format='pdf'))
                tree.append(index_nodes)

        # This is stolen from the HTML builder's prepare_writing function
        self.domain_indices = []
        # html_domain_indices can be False/True or a list of index names
        indices_config = self.config.pdf_domain_indices
        if indices_config and hasattr(self.env, 'domains'):
            for domain in self.env.domains.values():
                for indexcls in domain.indices:
                    indexname = '%s-%s' % (domain.name, indexcls.name)
                    if isinstance(indices_config, list):
                        if indexname not in indices_config:
                            continue
                    # deprecated config value
                    if indexname == 'py-modindex' and not self.config.pdf_use_modindex:
                        continue
                    content, collapse = indexcls(domain).generate()
                    if content:
                        self.domain_indices.append(
                            (indexname, indexcls, content, collapse)
                        )

        # self.domain_indices contains a list of indices to generate, like
        # this:
        # [('py-modindex',
        #    <class 'sphinx.domains.python.PythonModuleIndex'>,
        #   [(u'p', [[u'parrot', 0, 'test', u'module-parrot', 'Unix, Windows',
        #   '', 'Analyze and reanimate dead parrots.']])], True)]

        # Now this in the HTML builder is passed onto write_domain_indices.
        # We handle it right here

        for indexname, indexcls, content, collapse in self.domain_indices:
            # In HTML this is handled with a Jinja template, domainindex.html
            # We have to generate docutils stuff right here in the same way.
            self.sphinx_logger.info(' ' + indexname)

            output = ['DUMMY', '=====', '', '.. _modindex:\n\n']
            t = indexcls.localname
            t += '\n' + '=' * len(t) + '\n'
            output.append(t)

            for letter, entries in content:
                output.append('.. cssclass:: heading4\n\n%s\n\n' % letter)
                for (
                    name,
                    grouptype,
                    page,
                    anchor,
                    extra,
                    qualifier,
                    description,
                ) in entries:
                    if qualifier:
                        q = '[%s]' % qualifier
                    else:
                        q = ''

                    if extra:
                        e = '(%s)' % extra
                    else:
                        e = ''
                    output.append('`%s <#%s>`_ %s %s' % (name, anchor, e, q))
                    output.append('    %s' % description)
                output.append('')

            dt = docutils.core.publish_doctree('\n'.join(output))[1:]
            dt.insert(0, nodes.raw(text='OddPageBreak twoColumn', format='pdf'))
            tree.extend(dt)

        if appendices:
            tree.append(
                nodes.raw(text='OddPageBreak %s' % self.page_template, format='pdf')
            )
            self.sphinx_logger.info('')
            self.sphinx_logger.info('adding appendixes...')
            for docname in appendices:
                self.sphinx_logger.info(darkgreen(docname) + " ")
                appendix = self.env.get_doctree(docname)
                appendix['docname'] = docname
                tree.append(appendix)
            self.sphinx_logger.info('done')

        # Replace Sphinx's HighlightLanguageTransform with our own for sphinx version between 1.8.0 & less than 2.0.0 as
        # Sphinx's HighlightLanguageTransform breaks linenothreshold setting in the highlight directive (See issue #721)
        # This code can be removed when we drop support for Python 2
        if sphinx.__version__ > '1.7.9' and sphinx.__version__ < '2.0.0':
            for i in range(len(self.env.app.registry.post_transforms)):
                if (
                    self.env.app.registry.post_transforms[i].__name__
                    == 'HighlightLanguageTransform'
                ):
                    self.env.app.registry.post_transforms[
                        i
                    ] = HighlightLanguageTransform
                    break

        self.sphinx_logger.info("resolving references...")
        self.env.resolve_references(tree, docname, self)

        for pendingnode in tree.traverse(addnodes.pending_xref):
            # This needs work, need to keep track of all targets
            # so I don't replace and create hanging refs, which
            # crash
            if (
                pendingnode.get('reftarget', None) == 'genindex'
                and self.config.pdf_use_index
            ):
                pendingnode.replace_self(
                    nodes.reference(
                        text=pendingnode.astext(), refuri=pendingnode['reftarget']
                    )
                )
            # FIXME: probably need to handle dangling links to domain-specific indexes
            else:
                # FIXME: This is from the LaTeX builder and I still don't understand it
                # well, and doesn't seem to work

                # resolve :ref:s to distant tex files -- we can't add a cross-reference,
                # but append the document name
                docname = pendingnode['refdocname']
                sectname = pendingnode['refsectname']
                newnodes = [nodes.emphasis(sectname, sectname)]
                for subdir, title in self.titles:
                    if docname.startswith(subdir):
                        newnodes.append(nodes.Text(_(' (in '), _(' (in ')))
                        newnodes.append(nodes.emphasis(title, title))
                        newnodes.append(nodes.Text(')', ')'))
                        break
                else:
                    pass
                pendingnode.replace_self(newnodes)
            # else:
            # pass
        return tree

    def get_target_uri(self, docname, typ=None):
        # print 'GTU',docname,typ
        # FIXME: production lists are not supported yet!
        if typ == 'token':
            # token references are always inside production lists and must be
            # replaced by \token{} in LaTeX
            return '@token'
        if docname not in self.docnames:

            # It can be a 'main' document:
            for doc in self.document_data:
                if doc[0] == docname:
                    return "pdf:" + doc[1] + '.pdf'
            # It can be in some other document's toctree
            for indexname, toctree in self.env.toctree_includes.items():
                if docname in toctree:
                    for doc in self.document_data:
                        if doc[0] == indexname:
                            return "pdf:" + doc[1] + '.pdf'
            # No idea
            raise NoUri
        else:  # Local link
            return ""

    def get_relative_uri(self, from_, to, typ=None):
        # ignore source path
        return self.get_target_uri(to, typ)

    def get_outdated_docs(self):
        for docname in self.env.found_docs:
            if docname not in self.env.all_docs:
                yield docname
                continue
            targetname = os.path.join(self.outdir, docname + self.out_suffix)
            try:
                targetmtime = os.path.getmtime(targetname)
            except Exception:
                targetmtime = 0
            try:
                srcmtime = os.path.getmtime(self.env.doc2path(docname))
                if srcmtime > targetmtime:
                    yield docname
            except EnvironmentError:
                # source doesn't exist anymore
                pass


def genindex_nodes(genindexentries):
    indexlabel = _('Index')
    indexunder = '=' * len(indexlabel)
    output = ['DUMMY', '=====', '.. _genindex:\n\n', indexlabel, indexunder, '']

    for key, entries in genindexentries:
        output.append('.. cssclass:: heading4\n\n%s\n\n' % key)  # initial
        for entryname, entryvalue in entries:
            links, subitems = entryvalue[0:2]
            if links:
                output.append('`%s <#%s>`_' % (entryname, nodes.make_id(links[0][1])))
                for i, link in enumerate(links[1:]):
                    output[-1] += ' `[%s] <#%s>`_ ' % (i + 1, nodes.make_id(link[1]))
                output.append('')
            else:
                output.append(entryname)
            if subitems:
                for subentryname, subentrylinks in subitems:
                    if subentrylinks:
                        output.append(
                            '    `%s <%s>`_' % (subentryname, subentrylinks[0])
                        )
                        for i, link in enumerate(subentrylinks[1:]):
                            output[-1] += ' `[%s] <%s>`_ ' % (i + 1, link)
                        output.append('')
                    else:
                        output.append(subentryname)
                        output.append('')

    doctree = docutils.core.publish_doctree('\n'.join(output))
    return doctree[1]


class PDFContents(Contents):

    # Mostly copied from Docutils' Contents transformation

    def build_contents(self, node, level=0):
        level += 1
        sections = []
        # Replaced this with the for below to make it work for Sphinx
        # trees.

        # sections = [sect for sect in node if isinstance(sect, nodes.section)]
        for sect in node:
            if isinstance(sect, nodes.compound):
                for sect2 in sect:
                    if isinstance(sect2, addnodes.start_of_file):
                        for sect3 in sect2:
                            if isinstance(sect3, nodes.section):
                                sections.append(sect3)
            elif isinstance(sect, nodes.section):
                sections.append(sect)
        entries = []
        # FIXME: depth should be taken from :maxdepth: (Issue 320)
        depth = self.toc_depth
        for section in sections:
            title = section[0]
            auto = title.get('auto')  # May be set by SectNum.
            entrytext = self.copy_and_filter(title)
            reference = nodes.reference('', '', refid=section['ids'][0], *entrytext)
            ref_id = self.document.set_id(reference)
            entry = nodes.paragraph('', '', reference)
            item = nodes.list_item('', entry)
            if (
                self.backlinks in ('entry', 'top')
                and title.next_node(nodes.reference) is None
            ):
                if self.backlinks == 'entry':
                    title['refid'] = ref_id
                elif self.backlinks == 'top':
                    title['refid'] = self.toc_id
            if level < depth:
                subsects = self.build_contents(section, level)
                item += subsects
            entries.append(item)
        if entries:
            contents = nodes.bullet_list('', *entries)
            if auto:
                contents['classes'].append('auto-toc')
            return contents
        else:
            return []


class PDFWriter(writers.Writer):
    def __init__(
        self,
        builder,
        stylesheets,
        language,
        breaklevel=0,
        breakside='any',
        fontpath=[],
        fitmode='shrink',
        compressed=False,
        inline_footnotes=False,
        splittables=True,
        srcdir='.',
        default_dpi=300,
        page_template='decoratedPage',
        invariant=False,
        real_footnotes=False,
        use_toc=True,
        use_coverpage=True,
        toc_depth=9999,
        use_numbered_links=False,
        fit_background_mode="scale",
        section_header_depth=2,
        baseurl=urlunparse(['file', os.getcwd() + os.sep, '', '', '', '']),
        style_path=None,
        repeat_table_rows=False,
        smartquotes='0',
        config={},
    ):
        writers.Writer.__init__(self)
        self.builder = builder
        self.output = ''
        self.stylesheets = stylesheets
        self.__language = language
        self.breaklevel = int(breaklevel)
        self.breakside = breakside
        self.fontpath = fontpath
        self.fitmode = fitmode
        self.compressed = compressed
        self.inline_footnotes = inline_footnotes
        self.splittables = splittables
        self.highlightlang = builder.config.highlight_language
        self.srcdir = srcdir
        self.config = config
        self.default_dpi = default_dpi
        self.page_template = page_template
        self.invariant = invariant
        self.real_footnotes = real_footnotes
        self.use_toc = use_toc
        self.use_coverpage = use_coverpage
        self.toc_depth = toc_depth
        self.use_numbered_links = use_numbered_links
        self.fit_background_mode = fit_background_mode
        self.section_header_depth = section_header_depth
        self.repeat_table_rows = repeat_table_rows
        self.smartquotes = smartquotes
        self.baseurl = baseurl
        if hasattr(sys, 'frozen'):
            self.PATH = os.path.abspath(os.path.dirname(sys.executable))
        else:
            self.PATH = os.path.abspath(os.path.dirname(__file__))
        if style_path:
            self.style_path = style_path
        else:
            self.style_path = [self.srcdir]

    supported = 'pdf'
    config_section = 'pdf writer'
    config_section_dependencies = ('writers',)

    def translate(self):
        visitor = PDFTranslator(self.document, self.builder)
        self.document.walkabout(visitor)
        lang = self.config.language or 'en'
        langmod = get_language_available(lang)[2]
        self.docutils_languages = {lang: langmod}

        # Generate Contents topic manually
        if self.use_toc:
            contents = nodes.topic(classes=['contents'])
            contents += nodes.title('')
            contents[0] += nodes.Text(langmod.labels['contents'])
            contents['ids'] = ['Contents']
            pending = nodes.topic()
            contents.append(pending)
            pending.details = {}
            self.document.insert(
                0, nodes.raw(text='SetPageCounter 1 arabic', format='pdf')
            )
            self.document.insert(
                0, nodes.raw(text='OddPageBreak %s' % self.page_template, format='pdf')
            )
            self.document.insert(0, contents)
            self.document.insert(
                0, nodes.raw(text='SetPageCounter 1 lowerroman', format='pdf')
            )
            contTrans = PDFContents(self.document)
            contTrans.toc_depth = self.toc_depth
            contTrans.startnode = pending
            contTrans.apply()

        if self.use_coverpage:
            # Generate cover page

            # FIXME: duplicate from createpdf, refactor!
            # Add the Sphinx template paths
            def add_template_path(path):
                return os.path.join(self.srcdir, path)

            jinja_env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(
                    [
                        self.srcdir,
                        os.path.expanduser('~/.rst2pdf'),
                        os.path.join(self.PATH, 'templates'),
                    ]
                    + list(map(add_template_path, self.config.templates_path))
                ),
                autoescape=jinja2.select_autoescape(['html', 'xml']),
            )

            try:
                template = jinja_env.get_template(self.config.pdf_cover_template)
            except jinja2.TemplateNotFound:
                log.error(
                    "Can't find cover template %s, using default"
                    % self.config.pdf_cover_template
                )
                template = jinja_env.get_template('sphinxcover.tmpl')

            # This is what's used in the python docs because
            # Latex does a manual linebreak. This sucks.
            authors = self.document.settings.author.split('\\')

            # Honour the "today" config setting
            if self.config.today:
                date = self.config.today
            else:
                date = time.strftime(self.config.today_fmt or _('%B %d, %Y'))

            # Feed data to the template, get restructured text.
            cover_text = template.render(
                title=self.document.settings.title or visitor.elements['title'],
                subtitle='%s %s' % (self.config.subtitle_prefix, self.config.version),
                authors=authors,
                date=date,
            )

            cover_tree = docutils.core.publish_doctree(cover_text)
            self.document.insert(0, cover_tree)

        sio = BytesIO()

        if self.invariant:
            createpdf.patch_PDFDate()
            createpdf.patch_digester()

        createpdf.RstToPdf(
            sphinx=True,
            stylesheets=self.stylesheets,
            language=self.__language,
            breaklevel=self.breaklevel,
            breakside=self.breakside,
            fit_mode=self.fitmode,
            font_path=self.fontpath,
            inline_footnotes=self.inline_footnotes,
            highlightlang=self.highlightlang,
            splittables=self.splittables,
            style_path=self.style_path,
            repeat_table_rows=self.repeat_table_rows,
            basedir=self.srcdir,
            def_dpi=self.default_dpi,
            real_footnotes=self.real_footnotes,
            numbered_links=self.use_numbered_links,
            background_fit_mode=self.fit_background_mode,
            baseurl=self.baseurl,
            section_header_depth=self.section_header_depth,
            toc_depth=self.toc_depth,
            smarty=self.smartquotes,
        ).createPdf(doctree=self.document, output=sio, compressed=self.compressed)
        self.output = sio.getvalue()

    def supports(self, format):
        """This writer supports all format-specific elements."""
        return 1


class PDFTranslator(nodes.SparseNodeVisitor):
    def __init__(self, document, builder):
        nodes.NodeVisitor.__init__(self, document)
        self.builder = builder
        self.footnotestack = []
        self.curfilestack = []
        self.highlightlinenothreshold = 999999
        self.top_sectionlevel = 1
        self.footnotecounter = 1
        self.curfile = None
        self.footnotedict = {}
        self.this_is_the_title = True
        self.in_title = 0
        self.elements = {
            'title': document.settings.title,
        }
        self.highlightlang = builder.config.highlight_language

    def visit_document(self, node):
        self.curfilestack.append(node.get('docname', ''))
        self.footnotestack.append('')

    def visit_start_of_file(self, node):
        self.curfilestack.append(node['docname'])
        self.footnotestack.append(node['docname'])

    def depart_start_of_file(self, node):
        self.footnotestack.pop()
        self.curfilestack.pop()

    def visit_highlightlang(self, node):
        self.highlightlang = node['lang']
        self.highlightlinenothreshold = node['linenothreshold']
        raise nodes.SkipNode

    def visit_versionmodified(self, node):
        replacement = nodes.paragraph()
        replacement.extend(node.children)
        node.parent.replace(node, replacement)

    def depart_versionmodified(self, node):
        pass

    def visit_literal_block(self, node):
        if 'code' in node['classes']:  # Probably a processed code-block
            pass
        else:
            lang = lang_for_block(
                node.astext(), node.get('language', self.highlightlang)
            )
            content = node.astext().splitlines()
            if len(content) > self.highlightlinenothreshold or node.get(
                'linenos', False
            ):
                options = {'linenos': True}
            else:
                options = {}

            # FIXME: make tab width configurable
            content = [c.replace('\t', '        ') for c in content]
            replacement = nodes.literal_block()
            replacement.children = code_block.code_block_directive(
                name=None,
                arguments=[lang],
                options=options,
                content=content,
                lineno=False,
                content_offset=None,
                block_text=None,
                state=None,
                state_machine=None,
            )
            node.parent.replace(node, replacement)

    def visit_footnote(self, node):
        node['backrefs'] = [
            '%s_%s' % (self.footnotestack[-1], x) for x in node['backrefs']
        ]
        node['ids'] = ['%s_%s' % (self.footnotestack[-1], x) for x in node['ids']]
        node.children[0][0] = nodes.Text(str(self.footnotecounter))
        for id in node['backrefs']:
            try:
                fnr = self.footnotedict[id]
            except KeyError:
                pass
            else:
                fnr.children[0] = nodes.Text(str(self.footnotecounter))
        self.footnotedict[node['ids'][0]] = node
        self.footnotecounter += 1

    def visit_footnote_reference(self, node):
        node['ids'] = ['%s_%s' % (self.footnotestack[-1], x) for x in node['ids']]
        node['refid'] = '%s_%s' % (self.footnotestack[-1], node['refid'])
        self.footnotedict[node['ids'][0]] = node
        try:
            footnote = self.footnotedict[node['refid']]
        except KeyError:
            pass
        else:
            node.children[0] = nodes.Text(footnote.children[0][0])

    def visit_desc_annotation(self, node):
        pass

    def depart_desc_annotation(self, node):
        pass

    # This is for graphviz support
    def visit_graphviz(self, node):
        # Not neat, but I need to send self to my handlers
        node['builder'] = self

    def visit_Aanode(self, node):
        pass

    def depart_Aanode(self, node):
        pass

    def visit_productionlist(self, node):
        replacement = nodes.literal_block(classes=["code"])

        # Sphinx 8+ uses pre-formatted children in productionlist nodes (#13326)
        has_formatting = sphinx.version_info >= (8, 2)

        names = []
        for production in node:
            names.append(production['tokenname'])
        maxlen = max(len(name) for name in names)

        for production in node:
            if has_formatting:
                # New Sphinx/docutils: production children already contain formatted content
                # We need to reconstruct in the old format to avoid duplication
                replacement.children.extend(production.children)
            else:
                # Old Sphinx/docutils format, so we need to format manually
                if production['tokenname']:
                    lastname = production['tokenname'].ljust(maxlen)
                    n = nodes.strong()
                    n += nodes.Text(lastname)
                    replacement += n
                    replacement += nodes.Text(' ::=')
                else:
                    replacement += nodes.Text(' ' * (maxlen + 4))

                production.walkabout(self)
                replacement.children.extend(production.children)
                replacement += nodes.Text('\n')

        node.parent.replace(node, replacement)
        raise nodes.SkipNode

    def depart_productionlist(self, node):
        pass

    def visit_production(self, node):
        pass

    def depart_production(self, node):
        pass

    def visit_OddEvenNode(self, node):
        pass

    def depart_OddEvenNode(self, node):
        pass


class HighlightLanguageTransform(SphinxTransform):
    """
    This is a copy of Sphinx's HighlightLanguageTransform for use with Sphinx versions between 1.8.0 & less than 2.0.0
    as the Sphinx version of this class breaks the linenothreshold setting in the highlight directive (See issue #721).
    This code can be removed when we drop support for Python 2

    Apply highlight_language to all literal_block nodes.

    This refers both :confval:`highlight_language` setting and
    :rst:dir:`highlightlang` directive.

    After processing, this overridden transform DOES NOT REMOVE ``highlightlang`` node from doctree in order to allow
    pdfbuilder's visit_highlightlang to work as before.
    """

    default_priority = 400

    def apply(self):
        from sphinx.transforms.post_transforms.code import HighlightLanguageVisitor

        visitor = HighlightLanguageVisitor(
            self.document, self.config.highlight_language
        )
        self.document.walkabout(visitor)


# This is copied from sphinx.highlighting
def lang_for_block(source, lang):
    if lang in ('py', 'python'):
        if source.startswith('>>>'):
            # interactive session
            return 'pycon'
        else:
            # maybe Python -- try parsing it
            if try_parse(source):
                return 'python'
            else:  # Guess
                return lang_for_block(source, 'guess')
    elif lang in ('python3', 'py3') and source.startswith('>>>'):
        # for py3, recognize interactive sessions, but do not try parsing...
        return 'pycon3'
    elif lang == 'guess':
        try:
            # return 'python'
            lexer = guess_lexer(source)
            return lexer.aliases[0]
        except Exception:
            return None
    else:
        return lang


def try_parse(src):
    # Make sure it ends in a newline
    src += '\n'

    # Replace "..." by a mark which is also a valid python expression
    # (Note, the highlighter gets the original source, this is only done
    #  to allow "..." in code and still highlight it as Python code.)
    mark = "__highlighting__ellipsis__"
    src = src.replace("...", mark)

    # lines beginning with "..." are probably placeholders for suite
    src = re.sub(r"(?m)^(\s*)" + mark + "(.)", r"\1" + mark + r"# \2", src)

    if not isinstance(src, bytes):
        # Non-ASCII chars will only occur in string literals
        # and comments.  If we wanted to give them to the parser
        # correctly, we'd have to find out the correct source
        # encoding.  Since it may not even be given in a snippet,
        # just replace all non-ASCII characters.
        src = src.encode('ascii', 'replace')

    return True


def setup(app):
    app.add_builder(PDFBuilder)
    # PDF options
    app.add_config_value('pdf_documents', [], None)
    app.add_config_value('pdf_stylesheets', ['sphinx'], None)
    app.add_config_value('pdf_style_path', None, None)
    app.add_config_value('pdf_compressed', False, None)
    app.add_config_value('pdf_font_path', [], None)
    app.add_config_value('pdf_language', 'en_US', None)
    app.add_config_value('pdf_fit_mode', '', None),
    app.add_config_value('pdf_break_level', 0, None)
    app.add_config_value('pdf_inline_footnotes', True, None)
    app.add_config_value('pdf_verbosity', 0, None)
    app.add_config_value('pdf_use_index', True, None)
    app.add_config_value('pdf_domain_indices', True, None)
    app.add_config_value('pdf_use_modindex', True, None)
    app.add_config_value('pdf_use_coverpage', True, None)
    app.add_config_value('pdf_cover_template', 'sphinxcover.tmpl', None)
    app.add_config_value('pdf_appendices', [], None)
    app.add_config_value('pdf_splittables', True, None)
    app.add_config_value('pdf_repeat_table_rows', False, None)
    app.add_config_value('pdf_breakside', 'odd', None)
    app.add_config_value('pdf_default_dpi', 300, None)
    app.add_config_value('pdf_extensions', [], None)
    app.add_config_value('pdf_page_template', 'decoratedPage', None)
    app.add_config_value('pdf_invariant', False, None)
    app.add_config_value('pdf_real_footnotes', False, None)
    app.add_config_value('pdf_use_toc', True, None)
    app.add_config_value('pdf_toc_depth', 9999, None)
    app.add_config_value('pdf_use_numbered_links', False, None)
    app.add_config_value('pdf_fit_background_mode', "scale", None)
    app.add_config_value('pdf_smartquotes', 0, None)
    app.add_config_value('section_header_depth', 2, None)
    app.add_config_value(
        'pdf_baseurl', urlunparse(['file', os.getcwd() + os.sep, '', '', '', '']), None
    )
    app.add_config_value('subtitle_prefix', 'version', None)

    if len(app.config.pdf_documents) == 0:
        project_doc = app.config.project + ' Documentation'
        app.config.pdf_documents.append(
            (
                app.config.master_doc,
                app.config.project,
                project_doc,
                app.config.copyright,
                'manual',
            )
        )

    return {
        'version': rst2pdf.version,
        'parallel_read_safe': True,
        'parallel_write_safe': False,
    }



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms

from docutils.languages import get_language as get_language

from .log import log


def get_language_silent(lang):
    """Docutils get_language() with patches for older versions."""
    try:
        return get_language(lang)
    except TypeError as err:  # Docutils 0.8.1
        if 'get_language() takes exactly 2 arguments' in str(err):

            class SilentReporter(object):
                def warning(self, msg):
                    pass

            return get_language(lang, SilentReporter())
        raise  # re-raise any other TypeError
    except ImportError:  # Docutils < 0.8
        return get_language('en')


def get_language_available(lang):
    """Docutils get_language() also returning the available language."""
    module = get_language_silent(lang)
    docutils_lang = module.__name__.rsplit('.', 1)[-1]
    if docutils_lang == 'en' and docutils_lang != lang and '_' in lang:
        module = get_language_silent(lang.split('_', 1)[0])
        docutils_lang = module.__name__.rsplit('.', 1)[-1]
    if docutils_lang != lang:
        warn = (
            docutils_lang.split('_', 1)[0] == lang.split('_', 1)[0]
            and log.info
            or log.warning
        )
        warn(
            "Language '%s' not supported by Docutils,"
            " using '%s' instead." % (lang, docutils_lang)
        )
        if docutils_lang == 'en' and lang.split('_', 1)[0] != 'en':
            lang = 'en_US'
    return lang, docutils_lang, module



# -*- coding: utf-8 -*-
# See LICENSE.txt for licensing terms


def noop_directive(
    name,
    arguments,
    options,
    content,
    lineno,
    content_offset,
    block_text,
    state,
    state_machine,
):
    node_list = []
    return node_list


noop_directive.content = ["*"]



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

"""Convert older RSON stylesheets to YAML format

Run the script with this list of filenames to convert. It outputs to stdout, or
use the --save3 flag to have it create .yaml files
"""

import argparse
import json
import os

import yaml

from rst2pdf.dumpstyle import fixstyle
from rst2pdf.rson import loads as rloads


def main():
    # set up the command, optional --save parameter, and a list of paths
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        '--save',
        action='store_true',
        help='Save .yaml version of the file (rather than output to stdout)',
    )
    parser.add_argument(
        'paths',
        metavar='PATH',
        nargs='+',
        help='An RSON-formatted file to convert.',
    )
    args = parser.parse_args()

    # loop over the files
    for path in args.paths:
        # read rson from a file
        with open(path, 'rb') as fh:
            style_data = fixstyle(rloads(fh.read()))

        # output the style as json (already supported), then parse that
        json_style = json.dumps(style_data)
        reparsed_style = json.loads(json_style)

        yaml_style = yaml.dump(reparsed_style, default_flow_style=None)

        # output the yaml or save to a file
        if args.save:
            new_path = '.'.join((os.path.splitext(path)[0], 'yaml'))

            if os.path.exists(new_path):
                print("File " + new_path + " exists, cannot overwrite")
            else:
                print("Creating file " + new_path)
                with open(new_path, 'w') as file:
                    file.write(yaml_style)
        else:
            print("# " + os.path.splitext(path)[0])
            print("---")
            print(yaml_style)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2009 by Leandro Lucarella, Roberto Alsina
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from docutils.nodes import Element, literal_block
from docutils.parsers.rst import directives
from docutils.parsers import rst
from reportlab.graphics import renderPDF

from ..log import log

try:
    import aafigure
    import aafigure.pdf
except ImportError:
    aafigure = None

WARNED = False


class Aanode(Element):
    children = ()

    def __init__(self, content, options, rawsource='', *children, **attributes):
        self.content = content
        self.options = options
        Element.__init__(self, rawsource, *children, **attributes)

    def copy(self, **attributes):
        return Aanode(self.content, self.options, **self.attributes)

    def gen_flowable(self, style_options):
        options = dict(style_options)
        # explicit :option: always precedes
        options.update(self.options)
        visitor = aafigure.process(
            '\n'.join(self.content),
            aafigure.pdf.PDFOutputVisitor,
            options=options,
        )
        return renderPDF.GraphicsFlowable(visitor.drawing)


class Aafig(rst.Directive):
    """
    Directive to insert an ASCII art figure to be rendered by aafigure.
    """

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = dict(
        scale=float,
        line_width=float,
        background=str,
        foreground=str,
        fill=str,
        name=str,
        aspect=float,
        textual=directives.flag,
        proportional=directives.flag,
    )

    def run(self):
        global WARNED
        if 'textual' in self.options:
            self.options['textual'] = True
        if 'proportional' in self.options:
            self.options['proportional'] = True
        if aafigure is not None:
            return [Aanode(self.content, self.options)]
        if not WARNED:
            log.error(
                'To render the aafigure directive correctly, please install aafigure'
            )
            WARNED = True
        return [literal_block(text='\n'.join(self.content))]


directives.register_directive('aafig', Aafig)
directives.register_directive('aafigure', Aafig)



# -*- coding: utf-8 -*-

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['rst2pdf.pdfbuilder']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Foobar'
copyright = u'2009, Jason S'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '1.0.1'
# The full version, including alpha/beta/rc tags.
release = '1.0.1'


# -- Options for PDF output ----------------------------------------------------

# Grouping the document tree into PDF files. List of tuples
# (source start file, target name, title, author, options).
pdf_documents = [('index', u'index', u'index', u'lorenzo')]

# A comma-separated list of custom stylesheets. Example:
pdf_stylesheets = ['sphinx']

# Language to be used for hyphenation support
pdf_language = "en_US"

# Section level that forces a break page.
# For example: 1 means top-level sections start in a new page
# 0 means disabled
pdf_break_level = 1

# When a section starts in a new page, force it to be 'even', 'odd',
# or just use 'any'
pdf_breakside = 'odd'

# verbosity level. 0 1 or 2
pdf_verbosity = 0

# If false, no index is generated.
pdf_use_index = False

# If false, no coverpage is generated.
pdf_use_coverpage = False

pdf_page_template = 'customPage'
pdf_invariant = True
pdf_real_footnotes = True



# -*- coding: utf-8 -*-

"""A custom directive that allows alternative contents to be generated
on odd and even pages."""

from docutils.parsers import rst
from docutils.nodes import Admonition, Element
from docutils.parsers.rst import directives


class OddEvenNode(Admonition, Element):
    pass


class OddEven(rst.Directive):
    """A custom directive that allows alternative contents to be generated
    on odd and even pages. It can contain only two children, so use containers
    to group them. The first one is odd, the second is even."""

    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {}
    has_content = True

    def run(self):
        self.assert_has_content()
        node = OddEvenNode()
        self.state.nested_parse(self.content, self.content_offset, node)
        return [node]


directives.register_directive("oddeven", OddEven)



print(
    '''
This is a sample rst2pdf extension.

Because it is named 'sample.py' you can get rst2pdf to import it by
putting '-e sample' on the rst2pdf command line.

An extension is called after the command-line is parsed, and can
monkey-patch any necessary changes into rst2pdf.

An extension can live either in the extensions subdirectory, or
anywhere on the python path.
'''
)


def install(createpdf, options):
    """This function is called with an object with the createpdf
    module globals as attributes, and with the options from
    the command line parser.  This function does not have
    to exist, but must have the correct call signature if
    it does.
    """
